"""
    :codeauthor: :email:`Jeff Schroeder <jeffschroeder@computer.org>`
"""

import base64
from contextlib import contextmanager

import pytest
import salt.modules.kubernetesmod as kubernetesmod
import salt.states.kubernetes as kubernetes
import salt.utils.stringutils
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


@pytest.mark.skipif(
    kubernetesmod.HAS_LIBS is False,
    reason="Probably Kubernetes client lib is not installed. Skipping test_kubernetes.py",
)
class KubernetesTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.kubernetes
    """

    def setup_loader_modules(self):
        return {kubernetes: {"__env__": "base"}}

    @contextmanager
    def mock_func(self, func_name, return_value, test=False):
        """
        Mock any of the kubernetes state function return values and set
        the test options.
        """
        name = "kubernetes.{}".format(func_name)
        mocked = {name: MagicMock(return_value=return_value)}
        with patch.dict(kubernetes.__salt__, mocked) as patched:
            with patch.dict(kubernetes.__opts__, {"test": test}):
                yield patched

    def make_configmap(self, name, namespace="default", data=None):
        return self.make_ret_dict(
            kind="ConfigMap",
            name=name,
            namespace=namespace,
            data=data,
        )

    def make_secret(self, name, namespace="default", data=None):
        secret_data = self.make_ret_dict(
            kind="Secret",
            name=name,
            namespace=namespace,
            data=data,
        )
        # Base64 all of the values just like kubectl does
        for key, value in secret_data["data"].items():
            secret_data["data"][key] = base64.b64encode(
                salt.utils.stringutils.to_bytes(value)
            )

        return secret_data

    def make_node_labels(self, name="minikube"):
        return {
            "kubernetes.io/hostname": name,
            "beta.kubernetes.io/os": "linux",
            "beta.kubernetes.io/arch": "amd64",
            "failure-domain.beta.kubernetes.io/region": "us-west-1",
        }

    def make_node(self, name="minikube"):
        node_data = self.make_ret_dict(kind="Node", name="minikube")
        node_data.update(
            {
                "api_version": "v1",
                "kind": "Node",
                "metadata": {
                    "annotations": {"node.alpha.kubernetes.io/ttl": "0"},
                    "labels": self.make_node_labels(name=name),
                    "name": name,
                    "namespace": None,
                    "self_link": "/api/v1/nodes/{name}".format(name=name),
                    "uid": "7811b8ae-c1a1-11e7-a55a-0800279fb61e",
                },
                "spec": {"external_id": name},
                "status": {},
            }
        )
        return node_data

    def make_namespace(self, name="default"):
        namespace_data = self.make_ret_dict(kind="Namespace", name=name)
        del namespace_data["data"]
        namespace_data.update(
            {
                "status": {"phase": "Active"},
                "spec": {"finalizers": ["kubernetes"]},
                "metadata": {
                    "name": name,
                    "namespace": None,
                    "labels": None,
                    "self_link": "/api/v1/namespaces/{namespace}".format(
                        namespace=name,
                    ),
                    "annotations": None,
                    "uid": "752fceeb-c1a1-11e7-a55a-0800279fb61e",
                },
            }
        )
        return namespace_data

    def make_ret_dict(self, kind, name, namespace=None, data=None):
        """
        Make a minimal example configmap or secret for using in mocks
        """

        assert kind in ("Secret", "ConfigMap", "Namespace", "Node")

        if data is None:
            data = {}

        self_link = "/api/v1/namespaces/{namespace}/{kind}s/{name}".format(
            namespace=namespace,
            kind=kind.lower(),
            name=name,
        )

        return_data = {
            "kind": kind,
            "data": data,
            "api_version": "v1",
            "metadata": {
                "name": name,
                "labels": None,
                "namespace": namespace,
                "self_link": self_link,
                "annotations": {"kubernetes.io/change-cause": "salt-call state.apply"},
            },
        }
        return return_data

    def test_configmap_present__fail(self):
        error = kubernetes.configmap_present(
            name="testme",
            data={1: 1},
            source="salt://beyond/oblivion.jinja",
        )
        self.assertDictEqual(
            {
                "changes": {},
                "result": False,
                "name": "testme",
                "comment": "'source' cannot be used in combination with 'data'",
            },
            error,
        )

    def test_configmap_present__create_test_true(self):
        # Create a new configmap with test=True
        with self.mock_func("show_configmap", return_value=None, test=True):
            ret = kubernetes.configmap_present(
                name="example",
                data={"example.conf": "# empty config file"},
            )
            self.assertDictEqual(
                {
                    "comment": "The configmap is going to be created",
                    "changes": {},
                    "name": "example",
                    "result": None,
                },
                ret,
            )

    def test_configmap_present__create(self):
        # Create a new configmap
        with self.mock_func("show_configmap", return_value=None):
            cm = self.make_configmap(
                name="test",
                namespace="default",
                data={"foo": "bar"},
            )
            with self.mock_func("create_configmap", return_value=cm):
                actual = kubernetes.configmap_present(
                    name="test",
                    data={"foo": "bar"},
                )
                self.assertDictEqual(
                    {
                        "comment": "",
                        "changes": {"data": {"foo": "bar"}},
                        "name": "test",
                        "result": True,
                    },
                    actual,
                )

    def test_configmap_present__create_no_data(self):
        # Create a new configmap with no 'data' attribute
        with self.mock_func("show_configmap", return_value=None):
            cm = self.make_configmap(
                name="test",
                namespace="default",
            )
            with self.mock_func("create_configmap", return_value=cm):
                actual = kubernetes.configmap_present(name="test")
                self.assertDictEqual(
                    {
                        "comment": "",
                        "changes": {"data": {}},
                        "name": "test",
                        "result": True,
                    },
                    actual,
                )

    def test_configmap_present__replace_test_true(self):
        cm = self.make_configmap(
            name="settings",
            namespace="saltstack",
            data={"foobar.conf": "# Example configuration"},
        )
        with self.mock_func("show_configmap", return_value=cm, test=True):
            ret = kubernetes.configmap_present(
                name="settings",
                namespace="saltstack",
                data={"foobar.conf": "# Example configuration"},
            )
            self.assertDictEqual(
                {
                    "comment": "The configmap is going to be replaced",
                    "changes": {},
                    "name": "settings",
                    "result": None,
                },
                ret,
            )

    def test_configmap_present__replace(self):
        cm = self.make_configmap(name="settings", data={"action": "make=war"})
        # Replace an existing configmap
        with self.mock_func("show_configmap", return_value=cm):
            new_cm = cm.copy()
            new_cm.update({"data": {"action": "make=peace"}})
            with self.mock_func("replace_configmap", return_value=new_cm):
                actual = kubernetes.configmap_present(
                    name="settings",
                    data={"action": "make=peace"},
                )
                self.assertDictEqual(
                    {
                        "comment": (
                            "The configmap is already present. Forcing recreation"
                        ),
                        "changes": {"data": {"action": "make=peace"}},
                        "name": "settings",
                        "result": True,
                    },
                    actual,
                )

    def test_configmap_absent__noop_test_true(self):
        # Nothing to delete with test=True
        with self.mock_func("show_configmap", return_value=None, test=True):
            actual = kubernetes.configmap_absent(name="NOT_FOUND")
            self.assertDictEqual(
                {
                    "comment": "The configmap does not exist",
                    "changes": {},
                    "name": "NOT_FOUND",
                    "result": None,
                },
                actual,
            )

    def test_configmap_absent__test_true(self):
        # Configmap exists with test=True
        cm = self.make_configmap(name="deleteme", namespace="default")
        with self.mock_func("show_configmap", return_value=cm, test=True):
            actual = kubernetes.configmap_absent(name="deleteme")
            self.assertDictEqual(
                {
                    "comment": "The configmap is going to be deleted",
                    "changes": {},
                    "name": "deleteme",
                    "result": None,
                },
                actual,
            )

    def test_configmap_absent__noop(self):
        # Nothing to delete
        with self.mock_func("show_configmap", return_value=None):
            actual = kubernetes.configmap_absent(name="NOT_FOUND")
            self.assertDictEqual(
                {
                    "comment": "The configmap does not exist",
                    "changes": {},
                    "name": "NOT_FOUND",
                    "result": True,
                },
                actual,
            )

    def test_configmap_absent(self):
        # Configmap exists, delete it!
        cm = self.make_configmap(name="deleteme", namespace="default")
        with self.mock_func("show_configmap", return_value=cm):
            # The return from this module isn't used in the state
            with self.mock_func("delete_configmap", return_value={}):
                actual = kubernetes.configmap_absent(name="deleteme")
                self.assertDictEqual(
                    {
                        "comment": "ConfigMap deleted",
                        "changes": {
                            "kubernetes.configmap": {
                                "new": "absent",
                                "old": "present",
                            },
                        },
                        "name": "deleteme",
                        "result": True,
                    },
                    actual,
                )

    def test_secret_present__fail(self):
        actual = kubernetes.secret_present(
            name="sekret",
            data={"password": "monk3y"},
            source="salt://nope.jinja",
        )
        self.assertDictEqual(
            {
                "changes": {},
                "result": False,
                "name": "sekret",
                "comment": "'source' cannot be used in combination with 'data'",
            },
            actual,
        )

    def test_secret_present__exists_test_true(self):
        secret = self.make_secret(name="sekret")
        new_secret = secret.copy()
        new_secret.update({"data": {"password": "uncle"}})
        # Secret exists already and needs replacing with test=True
        with self.mock_func("show_secret", return_value=secret):
            with self.mock_func("replace_secret", return_value=new_secret, test=True):
                actual = kubernetes.secret_present(
                    name="sekret",
                    data={"password": "uncle"},
                )
                self.assertDictEqual(
                    {
                        "changes": {},
                        "result": None,
                        "name": "sekret",
                        "comment": "The secret is going to be replaced",
                    },
                    actual,
                )

    def test_secret_present__exists(self):
        # Secret exists and gets replaced
        secret = self.make_secret(name="sekret", data={"password": "booyah"})
        with self.mock_func("show_secret", return_value=secret):
            with self.mock_func("replace_secret", return_value=secret):
                actual = kubernetes.secret_present(
                    name="sekret",
                    data={"password": "booyah"},
                )
                self.assertDictEqual(
                    {
                        "changes": {"data": ["password"]},
                        "result": True,
                        "name": "sekret",
                        "comment": "The secret is already present. Forcing recreation",
                    },
                    actual,
                )

    def test_secret_present__create(self):
        # Secret exists and gets replaced
        secret = self.make_secret(name="sekret", data={"password": "booyah"})
        with self.mock_func("show_secret", return_value=None):
            with self.mock_func("create_secret", return_value=secret):
                actual = kubernetes.secret_present(
                    name="sekret",
                    data={"password": "booyah"},
                )
                self.assertDictEqual(
                    {
                        "changes": {"data": ["password"]},
                        "result": True,
                        "name": "sekret",
                        "comment": "",
                    },
                    actual,
                )

    def test_secret_present__create_no_data(self):
        # Secret exists and gets replaced
        secret = self.make_secret(name="sekret")
        with self.mock_func("show_secret", return_value=None):
            with self.mock_func("create_secret", return_value=secret):
                actual = kubernetes.secret_present(name="sekret")
                self.assertDictEqual(
                    {
                        "changes": {"data": []},
                        "result": True,
                        "name": "sekret",
                        "comment": "",
                    },
                    actual,
                )

    def test_secret_present__create_test_true(self):
        # Secret exists and gets replaced with test=True
        secret = self.make_secret(name="sekret")
        with self.mock_func("show_secret", return_value=None):
            with self.mock_func("create_secret", return_value=secret, test=True):
                actual = kubernetes.secret_present(name="sekret")
                self.assertDictEqual(
                    {
                        "changes": {},
                        "result": None,
                        "name": "sekret",
                        "comment": "The secret is going to be created",
                    },
                    actual,
                )

    def test_secret_absent__noop_test_true(self):
        with self.mock_func("show_secret", return_value=None, test=True):
            actual = kubernetes.secret_absent(name="sekret")
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": None,
                    "name": "sekret",
                    "comment": "The secret does not exist",
                },
                actual,
            )

    def test_secret_absent__noop(self):
        with self.mock_func("show_secret", return_value=None):
            actual = kubernetes.secret_absent(name="passwords")
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": True,
                    "name": "passwords",
                    "comment": "The secret does not exist",
                },
                actual,
            )

    def test_secret_absent__delete_test_true(self):
        secret = self.make_secret(name="credentials", data={"redis": "letmein"})
        with self.mock_func("show_secret", return_value=secret):
            with self.mock_func("delete_secret", return_value=secret, test=True):
                actual = kubernetes.secret_absent(name="credentials")
                self.assertDictEqual(
                    {
                        "changes": {},
                        "result": None,
                        "name": "credentials",
                        "comment": "The secret is going to be deleted",
                    },
                    actual,
                )

    def test_secret_absent__delete(self):
        secret = self.make_secret(name="foobar", data={"redis": "letmein"})
        deleted = {
            "status": None,
            "kind": "Secret",
            "code": None,
            "reason": None,
            "details": None,
            "message": None,
            "api_version": "v1",
            "metadata": {
                "self_link": "/api/v1/namespaces/default/secrets/foobar",
                "resource_version": "30292",
            },
        }
        with self.mock_func("show_secret", return_value=secret):
            with self.mock_func("delete_secret", return_value=deleted):
                actual = kubernetes.secret_absent(name="foobar")
                self.assertDictEqual(
                    {
                        "changes": {
                            "kubernetes.secret": {"new": "absent", "old": "present"},
                        },
                        "result": True,
                        "name": "foobar",
                        "comment": "Secret deleted",
                    },
                    actual,
                )

    def test_node_label_present__add_test_true(self):
        labels = self.make_node_labels()
        with self.mock_func("node_labels", return_value=labels, test=True):
            actual = kubernetes.node_label_present(
                name="com.zoo-animal",
                node="minikube",
                value="monkey",
            )
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": None,
                    "name": "com.zoo-animal",
                    "comment": "The label is going to be set",
                },
                actual,
            )

    def test_node_label_present__add(self):
        node_data = self.make_node()
        # Remove some of the defaults to make it simpler
        node_data["metadata"]["labels"] = {
            "beta.kubernetes.io/os": "linux",
        }
        labels = node_data["metadata"]["labels"]

        with self.mock_func("node_labels", return_value=labels):
            with self.mock_func("node_add_label", return_value=node_data):
                actual = kubernetes.node_label_present(
                    name="failure-domain.beta.kubernetes.io/zone",
                    node="minikube",
                    value="us-central1-a",
                )
                self.assertDictEqual(
                    {
                        "comment": "",
                        "changes": {
                            "minikube.failure-domain.beta.kubernetes.io/zone": {
                                "new": {
                                    "failure-domain.beta.kubernetes.io/zone": (
                                        "us-central1-a"
                                    ),
                                    "beta.kubernetes.io/os": "linux",
                                },
                                "old": {"beta.kubernetes.io/os": "linux"},
                            },
                        },
                        "name": "failure-domain.beta.kubernetes.io/zone",
                        "result": True,
                    },
                    actual,
                )

    def test_node_label_present__already_set(self):
        node_data = self.make_node()
        labels = node_data["metadata"]["labels"]
        with self.mock_func("node_labels", return_value=labels):
            with self.mock_func("node_add_label", return_value=node_data):
                actual = kubernetes.node_label_present(
                    name="failure-domain.beta.kubernetes.io/region",
                    node="minikube",
                    value="us-west-1",
                )
                self.assertDictEqual(
                    {
                        "changes": {},
                        "result": True,
                        "name": "failure-domain.beta.kubernetes.io/region",
                        "comment": (
                            "The label is already set and has the specified value"
                        ),
                    },
                    actual,
                )

    def test_node_label_present__update_test_true(self):
        node_data = self.make_node()
        labels = node_data["metadata"]["labels"]
        with self.mock_func("node_labels", return_value=labels):
            with self.mock_func("node_add_label", return_value=node_data, test=True):
                actual = kubernetes.node_label_present(
                    name="failure-domain.beta.kubernetes.io/region",
                    node="minikube",
                    value="us-east-1",
                )
                self.assertDictEqual(
                    {
                        "changes": {},
                        "result": None,
                        "name": "failure-domain.beta.kubernetes.io/region",
                        "comment": "The label is going to be updated",
                    },
                    actual,
                )

    def test_node_label_present__update(self):
        node_data = self.make_node()
        # Remove some of the defaults to make it simpler
        node_data["metadata"]["labels"] = {
            "failure-domain.beta.kubernetes.io/region": "us-west-1",
        }
        labels = node_data["metadata"]["labels"]
        with self.mock_func("node_labels", return_value=labels):
            with self.mock_func("node_add_label", return_value=node_data):
                actual = kubernetes.node_label_present(
                    name="failure-domain.beta.kubernetes.io/region",
                    node="minikube",
                    value="us-east-1",
                )
                self.assertDictEqual(
                    {
                        "changes": {
                            "minikube.failure-domain.beta.kubernetes.io/region": {
                                "new": {
                                    "failure-domain.beta.kubernetes.io/region": (
                                        "us-east-1"
                                    )
                                },
                                "old": {
                                    "failure-domain.beta.kubernetes.io/region": (
                                        "us-west-1"
                                    )
                                },
                            }
                        },
                        "result": True,
                        "name": "failure-domain.beta.kubernetes.io/region",
                        "comment": "The label is already set, changing the value",
                    },
                    actual,
                )

    def test_node_label_absent__noop_test_true(self):
        labels = self.make_node_labels()
        with self.mock_func("node_labels", return_value=labels, test=True):
            actual = kubernetes.node_label_absent(
                name="non-existent-label",
                node="minikube",
            )
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": None,
                    "name": "non-existent-label",
                    "comment": "The label does not exist",
                },
                actual,
            )

    def test_node_label_absent__noop(self):
        labels = self.make_node_labels()
        with self.mock_func("node_labels", return_value=labels):
            actual = kubernetes.node_label_absent(
                name="non-existent-label",
                node="minikube",
            )
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": True,
                    "name": "non-existent-label",
                    "comment": "The label does not exist",
                },
                actual,
            )

    def test_node_label_absent__delete_test_true(self):
        labels = self.make_node_labels()
        with self.mock_func("node_labels", return_value=labels, test=True):
            actual = kubernetes.node_label_absent(
                name="failure-domain.beta.kubernetes.io/region",
                node="minikube",
            )
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": None,
                    "name": "failure-domain.beta.kubernetes.io/region",
                    "comment": "The label is going to be deleted",
                },
                actual,
            )

    def test_node_label_absent__delete(self):
        node_data = self.make_node()
        labels = node_data["metadata"]["labels"].copy()

        node_data["metadata"]["labels"].pop("failure-domain.beta.kubernetes.io/region")

        with self.mock_func("node_labels", return_value=labels):
            with self.mock_func("node_remove_label", return_value=node_data):
                actual = kubernetes.node_label_absent(
                    name="failure-domain.beta.kubernetes.io/region",
                    node="minikube",
                )
                self.assertDictEqual(
                    {
                        "result": True,
                        "changes": {
                            "kubernetes.node_label": {
                                "new": "absent",
                                "old": "present",
                            }
                        },
                        "comment": "Label removed from node",
                        "name": "failure-domain.beta.kubernetes.io/region",
                    },
                    actual,
                )

    def test_namespace_present__create_test_true(self):
        with self.mock_func("show_namespace", return_value=None, test=True):
            actual = kubernetes.namespace_present(name="saltstack")
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": None,
                    "name": "saltstack",
                    "comment": "The namespace is going to be created",
                },
                actual,
            )

    def test_namespace_present__create(self):
        namespace_data = self.make_namespace(name="saltstack")
        with self.mock_func("show_namespace", return_value=None):
            with self.mock_func("create_namespace", return_value=namespace_data):
                actual = kubernetes.namespace_present(name="saltstack")
                self.assertDictEqual(
                    {
                        "changes": {"namespace": {"new": namespace_data, "old": {}}},
                        "result": True,
                        "name": "saltstack",
                        "comment": "",
                    },
                    actual,
                )

    def test_namespace_present__noop_test_true(self):
        namespace_data = self.make_namespace(name="saltstack")
        with self.mock_func("show_namespace", return_value=namespace_data, test=True):
            actual = kubernetes.namespace_present(name="saltstack")
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": None,
                    "name": "saltstack",
                    "comment": "The namespace already exists",
                },
                actual,
            )

    def test_namespace_present__noop(self):
        namespace_data = self.make_namespace(name="saltstack")
        with self.mock_func("show_namespace", return_value=namespace_data):
            actual = kubernetes.namespace_present(name="saltstack")
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": True,
                    "name": "saltstack",
                    "comment": "The namespace already exists",
                },
                actual,
            )

    def test_namespace_absent__noop_test_true(self):
        with self.mock_func("show_namespace", return_value=None, test=True):
            actual = kubernetes.namespace_absent(name="salt")
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": None,
                    "name": "salt",
                    "comment": "The namespace does not exist",
                },
                actual,
            )

    def test_namespace_absent__noop(self):
        with self.mock_func("show_namespace", return_value=None):
            actual = kubernetes.namespace_absent(name="salt")
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": True,
                    "name": "salt",
                    "comment": "The namespace does not exist",
                },
                actual,
            )

    def test_namespace_absent__delete_test_true(self):
        namespace_data = self.make_namespace(name="salt")
        with self.mock_func("show_namespace", return_value=namespace_data, test=True):
            actual = kubernetes.namespace_absent(name="salt")
            self.assertDictEqual(
                {
                    "changes": {},
                    "result": None,
                    "name": "salt",
                    "comment": "The namespace is going to be deleted",
                },
                actual,
            )

    def test_namespace_absent__delete_code_200(self):
        namespace_data = self.make_namespace(name="salt")
        deleted = namespace_data.copy()
        deleted["code"] = 200
        deleted.update({"code": 200, "message": None})
        with self.mock_func("show_namespace", return_value=namespace_data):
            with self.mock_func("delete_namespace", return_value=deleted):
                actual = kubernetes.namespace_absent(name="salt")
                self.assertDictEqual(
                    {
                        "changes": {
                            "kubernetes.namespace": {"new": "absent", "old": "present"}
                        },
                        "result": True,
                        "name": "salt",
                        "comment": "Terminating",
                    },
                    actual,
                )

    def test_namespace_absent__delete_status_terminating(self):
        namespace_data = self.make_namespace(name="salt")
        deleted = namespace_data.copy()
        deleted.update(
            {
                "code": None,
                "status": "Terminating namespace",
                "message": "Terminating this shizzzle yo",
            }
        )
        with self.mock_func("show_namespace", return_value=namespace_data):
            with self.mock_func("delete_namespace", return_value=deleted):
                actual = kubernetes.namespace_absent(name="salt")
                self.assertDictEqual(
                    {
                        "changes": {
                            "kubernetes.namespace": {"new": "absent", "old": "present"}
                        },
                        "result": True,
                        "name": "salt",
                        "comment": "Terminating this shizzzle yo",
                    },
                    actual,
                )

    def test_namespace_absent__delete_status_phase_terminating(self):
        # This is what kubernetes 1.8.0 looks like when deleting namespaces
        namespace_data = self.make_namespace(name="salt")
        deleted = namespace_data.copy()
        deleted.update(
            {"code": None, "message": None, "status": {"phase": "Terminating"}}
        )
        with self.mock_func("show_namespace", return_value=namespace_data):
            with self.mock_func("delete_namespace", return_value=deleted):
                actual = kubernetes.namespace_absent(name="salt")
                self.assertDictEqual(
                    {
                        "changes": {
                            "kubernetes.namespace": {"new": "absent", "old": "present"}
                        },
                        "result": True,
                        "name": "salt",
                        "comment": "Terminating",
                    },
                    actual,
                )

    def test_namespace_absent__delete_error(self):
        namespace_data = self.make_namespace(name="salt")
        deleted = namespace_data.copy()
        deleted.update({"code": 418, "message": "I' a teapot!", "status": None})
        with self.mock_func("show_namespace", return_value=namespace_data):
            with self.mock_func("delete_namespace", return_value=deleted):
                actual = kubernetes.namespace_absent(name="salt")
                self.assertDictEqual(
                    {
                        "changes": {},
                        "result": False,
                        "name": "salt",
                        "comment": "Something went wrong, response: {}".format(
                            deleted,
                        ),
                    },
                    actual,
                )



"""
    :codeauthor: Mike Place <mp@saltstack.com>
"""

import io

import salt.modules.cron as cronmod
import salt.states.cron as cron
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class CronTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        loader_gloabals = {
            "__low__": {"__id__": "noo"},
            "__grains__": {"os": "Debian", "os_family": "Debian"},
            "__opts__": {"test": False},
            "__salt__": {
                "cmd.run_all": MagicMock(
                    return_value={"pid": 5, "retcode": 0, "stderr": "", "stdout": ""}
                ),
                "cron.list_tab": cronmod.list_tab,
                "cron.rm_job": cronmod.rm_job,
                "cron.set_job": cronmod.set_job,
            },
        }
        return {cron: loader_gloabals, cronmod: loader_gloabals}

    def get_crontab(self, *args, **kwargs):
        return self._crontab.getvalue()

    def set_crontab(self, data):
        self._crontab.seek(0)
        self._crontab.truncate(0)
        self._crontab.write(data)

    def write_crontab(self, *args, **kw):
        self.set_crontab("\n".join([a.strip() for a in args[1]]))
        return {
            "retcode": False,
        }

    def setUp(self):
        super().setUp()
        self._crontab = io.StringIO()
        self.addCleanup(delattr, self, "_crontab")
        self.set_crontab("")
        for mod, mock in (
            ("salt.modules.cron.raw_cron", MagicMock(side_effect=self.get_crontab)),
            (
                "salt.modules.cron._write_cron_lines",
                MagicMock(side_effect=self.write_crontab),
            ),
        ):
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)
        self.addCleanup(self.set_crontab, "")

    def test_present(self):
        cron.present(name="foo", hour="1", identifier="1", user="root")
        self.assertMultiLineEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:1\n"
            "* 1 * * * foo",
        )
        cron.present(name="foo", hour="2", identifier="1", user="root")
        self.assertMultiLineEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:1\n"
            "* 2 * * * foo",
        )
        cron.present(
            name="cmd1",
            minute="0",
            comment="Commented cron job",
            commented=True,
            identifier="commented_1",
            user="root",
        )
        self.assertMultiLineEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:1\n"
            "* 2 * * * foo\n"
            "# Commented cron job SALT_CRON_IDENTIFIER:commented_1\n"
            "#DISABLED#0 * * * * cmd1",
        )
        cron.present(name="foo", hour="2", identifier="2", user="root")
        self.assertMultiLineEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:1\n"
            "* 2 * * * foo\n"
            "# Commented cron job SALT_CRON_IDENTIFIER:commented_1\n"
            "#DISABLED#0 * * * * cmd1\n"
            "# SALT_CRON_IDENTIFIER:2\n"
            "* 2 * * * foo",
        )
        cron.present(name="cmd2", commented=True, identifier="commented_2", user="root")
        self.assertMultiLineEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:1\n"
            "* 2 * * * foo\n"
            "# Commented cron job SALT_CRON_IDENTIFIER:commented_1\n"
            "#DISABLED#0 * * * * cmd1\n"
            "# SALT_CRON_IDENTIFIER:2\n"
            "* 2 * * * foo\n"
            "# SALT_CRON_IDENTIFIER:commented_2\n"
            "#DISABLED#* * * * * cmd2",
        )
        self.set_crontab(
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:1\n"
            "* 2 * * * foo\n"
            "# SALT_CRON_IDENTIFIER:2\n"
            "* 2 * * * foo\n"
            "* 2 * * * foo\n"
        )
        cron.present(name="foo", hour="2", user="root", identifier=None)
        self.assertEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:1\n"
            "* 2 * * * foo\n"
            "# SALT_CRON_IDENTIFIER:2\n"
            "* 2 * * * foo\n"
            "* 2 * * * foo",
        )

    def test_remove(self):
        with patch.dict(cron.__opts__, {"test": True}):
            self.set_crontab(
                "# Lines below here are managed by Salt, do not edit\n"
                "# SALT_CRON_IDENTIFIER:1\n"
                "* 1 * * * foo"
            )
            result = cron.absent(name="bar", identifier="1")
            self.assertEqual(
                result,
                {
                    "changes": {},
                    "comment": "Cron bar is set to be removed",
                    "name": "bar",
                    "result": None,
                },
            )
            self.assertEqual(
                self.get_crontab(),
                "# Lines below here are managed by Salt, do not edit\n"
                "# SALT_CRON_IDENTIFIER:1\n"
                "* 1 * * * foo",
            )
        with patch.dict(cron.__opts__, {"test": False}):
            self.set_crontab(
                "# Lines below here are managed by Salt, do not edit\n"
                "# SALT_CRON_IDENTIFIER:1\n"
                "* 1 * * * foo"
            )
            cron.absent(name="bar", identifier="1")
            self.assertEqual(
                self.get_crontab(),
                "# Lines below here are managed by Salt, do not edit",
            )
            self.set_crontab(
                "# Lines below here are managed by Salt, do not edit\n* * * * * foo"
            )
            cron.absent(name="bar", identifier="1")
            self.assertEqual(
                self.get_crontab(),
                "# Lines below here are managed by Salt, do not edit\n* * * * * foo",
            )
            # old behavior, do not remove with identifier set and
            # even if command match !
            self.set_crontab(
                "# Lines below here are managed by Salt, do not edit\n* * * * * foo"
            )
            cron.absent(name="foo", identifier="1")
            self.assertEqual(
                self.get_crontab(),
                "# Lines below here are managed by Salt, do not edit",
            )
            # old behavior, remove if no identifier and command match
            self.set_crontab(
                "# Lines below here are managed by Salt, do not edit\n* * * * * foo"
            )
            cron.absent(name="foo")
            self.assertEqual(
                self.get_crontab(),
                "# Lines below here are managed by Salt, do not edit",
            )

    def test_multiline_comments_are_updated(self):
        self.set_crontab(
            "# Lines below here are managed by Salt, do not edit\n"
            "# First crontab - single line comment SALT_CRON_IDENTIFIER:1\n"
            "* 1 * * * foo"
        )
        cron.present(
            name="foo",
            hour="1",
            comment="First crontab\nfirst multi-line comment\n",
            identifier="1",
            user="root",
        )
        cron.present(
            name="foo",
            hour="1",
            comment="First crontab\nsecond multi-line comment\n",
            identifier="1",
            user="root",
        )
        cron.present(
            name="foo",
            hour="1",
            comment="Second crontab\nmulti-line comment\n",
            identifier="2",
            user="root",
        )
        self.assertEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# First crontab\n"
            "# second multi-line comment\n"
            "#  SALT_CRON_IDENTIFIER:1\n"
            "* 1 * * * foo\n"
            "# Second crontab\n"
            "# multi-line comment\n"
            "#  SALT_CRON_IDENTIFIER:2\n"
            "* 1 * * * foo",
        )

    def test_existing_unmanaged_jobs_are_made_managed(self):
        self.set_crontab(
            "# Lines below here are managed by Salt, do not edit\n0 2 * * * foo"
        )
        ret = cron._check_cron("root", "foo", hour="2", minute="0")
        self.assertEqual(ret, "present")
        ret = cron.present("foo", "root", minute="0", hour="2")
        self.assertEqual(ret["changes"], {"root": "foo"})
        self.assertEqual(ret["comment"], "Cron foo updated")
        self.assertEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:foo\n"
            "0 2 * * * foo",
        )
        ret = cron.present("foo", "root", minute="0", hour="2")
        self.assertEqual(ret["changes"], {})
        self.assertEqual(ret["comment"], "Cron foo already present")

    def test_existing_noid_jobs_are_updated_with_identifier(self):
        self.set_crontab(
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:NO ID SET\n"
            "1 * * * * foo"
        )
        ret = cron._check_cron("root", "foo", minute=1)
        self.assertEqual(ret, "present")
        ret = cron.present("foo", "root", minute=1)
        self.assertEqual(ret["changes"], {"root": "foo"})
        self.assertEqual(ret["comment"], "Cron foo updated")
        self.assertEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:foo\n"
            "1 * * * * foo",
        )

    def test_existing_duplicate_unmanaged_jobs_are_merged_and_given_id(self):
        self.set_crontab(
            "# Lines below here are managed by Salt, do not edit\n"
            "0 2 * * * foo\n"
            "0 2 * * * foo"
        )
        ret = cron._check_cron("root", "foo", hour="2", minute="0")
        self.assertEqual(ret, "present")
        ret = cron.present("foo", "root", minute="0", hour="2")
        self.assertEqual(ret["changes"], {"root": "foo"})
        self.assertEqual(ret["comment"], "Cron foo updated")
        self.assertEqual(
            self.get_crontab(),
            "# Lines below here are managed by Salt, do not edit\n"
            "# SALT_CRON_IDENTIFIER:foo\n"
            "0 2 * * * foo",
        )
        ret = cron.present("foo", "root", minute="0", hour="2")
        self.assertEqual(ret["changes"], {})
        self.assertEqual(ret["comment"], "Cron foo already present")



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import os.path

import salt.modules.key as key
import salt.utils.crypt
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class KeyTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.key
    """

    def setup_loader_modules(self):
        return {key: {}}

    def test_finger(self):
        """
        Test for finger
        """
        with patch.object(os.path, "join", return_value="A"):
            with patch.object(salt.utils.crypt, "pem_finger", return_value="A"):
                with patch.dict(
                    key.__opts__,
                    {"pki_dir": MagicMock(return_value="A"), "hash_type": "sha256"},
                ):
                    self.assertEqual(key.finger(), "A")

    def test_finger_master(self):
        """
        Test for finger
        """
        with patch.object(os.path, "join", return_value="A"):
            with patch.object(salt.utils.crypt, "pem_finger", return_value="A"):
                with patch.dict(key.__opts__, {"pki_dir": "A", "hash_type": "sha256"}):
                    self.assertEqual(key.finger_master(), "A")



import salt.modules.win_certutil as certutil
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class CertUtilTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {certutil: {}}

    def test_get_serial(self):
        """
        Test getting the serial number from a certificate
        """
        expected = "180720d39cd2db3244ba037417241e90"
        mock = MagicMock(
            return_value=(
                "CertInfo\r\n"
                "Cert Serial Number: 180720d39cd2db3244ba037417241e90\r\n"
                "\r\n"
                "OtherStuff"
            )
        )
        with patch.dict(certutil.__salt__, {"cmd.run": mock}):
            out = certutil.get_cert_serial("/path/to/cert.cer")
            mock.assert_called_once_with(
                "certutil.exe -silent -verify /path/to/cert.cer"
            )
            self.assertEqual(expected, out)

    def test_get_serials(self):
        """
        Test getting all the serial numbers from a store
        """
        expected = [
            "180720d39cd2db3244ba037417241e90",
            "1768ac4e5b72bf1d0df0df118b34b959",
        ]
        mock = MagicMock(
            return_value=(
                "CertInfo\r\n"
                "================ Certificate 0 ================\r\n"
                "Serial Number: 180720d39cd2db3244ba037417241e90\r\n"
                "OtherStuff\r\n"
                "\r\n"
                "================ Certificate 1 ================\r\n"
                "Serial Number: 1768ac4e5b72bf1d0df0df118b34b959\r\n"
                "OtherStuff"
            )
        )
        with patch.dict(certutil.__salt__, {"cmd.run": mock}):
            out = certutil.get_stored_cert_serials("TrustedPublisher")
            mock.assert_called_once_with("certutil.exe -store TrustedPublisher")
            self.assertEqual(expected, out)

    def test_add_store(self):
        """
        Test adding a certificate to a specific store
        """
        cmd_mock = MagicMock(
            return_value=(
                "CertInfo\r\n"
                "================ Certificate 0 ================\r\n"
                "Serial Number: 180720d39cd2db3244ba037417241e90\r\n"
                "OtherStuff"
            )
        )
        cache_mock = MagicMock(return_value="/tmp/cert.cer")
        with patch.dict(
            certutil.__salt__, {"cmd.run": cmd_mock, "cp.cache_file": cache_mock}
        ):
            certutil.add_store("salt://path/to/file", "TrustedPublisher")
            cmd_mock.assert_called_once_with(
                "certutil.exe -addstore TrustedPublisher /tmp/cert.cer"
            )
            cache_mock.assert_called_once_with("salt://path/to/file", "base")

    def test_del_store(self):
        """
        Test removing a certificate to a specific store
        """
        with patch("salt.modules.win_certutil.get_cert_serial") as cert_serial_mock:
            cmd_mock = MagicMock(
                return_value=(
                    "CertInfo\r\n"
                    "================ Certificate 0 ================\r\n"
                    "Serial Number: 180720d39cd2db3244ba037417241e90\r\n"
                    "OtherStuff"
                )
            )
            cache_mock = MagicMock(return_value="/tmp/cert.cer")
            cert_serial_mock.return_value = "ABCDEF"
            with patch.dict(
                certutil.__salt__, {"cmd.run": cmd_mock, "cp.cache_file": cache_mock}
            ):
                certutil.del_store("salt://path/to/file", "TrustedPublisher")
                cmd_mock.assert_called_once_with(
                    "certutil.exe -delstore TrustedPublisher ABCDEF"
                )
                cache_mock.assert_called_once_with("salt://path/to/file", "base")



"""
Module for Solaris 10's zonecfg

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:platform:      OmniOS,OpenIndiana,SmartOS,OpenSolaris,Solaris 10
:depend:        salt.modules.file

.. versionadded:: 2017.7.0

.. warning::
    Oracle Solaris 11's zonecfg is not supported by this module!
"""

import logging
import re

import salt.utils.args
import salt.utils.data
import salt.utils.decorators
import salt.utils.files
import salt.utils.path
from salt.utils.odict import OrderedDict

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "zonecfg"

# Function aliases
__func_alias__ = {"import_": "import"}

# Global data
_zonecfg_info_resources = [
    "rctl",
    "net",
    "fs",
    "device",
    "dedicated-cpu",
    "dataset",
    "attr",
]

_zonecfg_info_resources_calculated = [
    "capped-cpu",
    "capped-memory",
]

_zonecfg_resource_setters = {
    "fs": ["dir", "special", "raw", "type", "options"],
    "net": [
        "address",
        "allowed-address",
        "global-nic",
        "mac-addr",
        "physical",
        "property",
        "vlan-id defrouter",
    ],
    "device": ["match", "property"],
    "rctl": ["name", "value"],
    "attr": ["name", "type", "value"],
    "dataset": ["name"],
    "dedicated-cpu": ["ncpus", "importance"],
    "capped-cpu": ["ncpus"],
    "capped-memory": ["physical", "swap", "locked"],
    "admin": ["user", "auths"],
}

_zonecfg_resource_default_selectors = {
    "fs": "dir",
    "net": "mac-addr",
    "device": "match",
    "rctl": "name",
    "attr": "name",
    "dataset": "name",
    "admin": "user",
}


@salt.utils.decorators.memoize
def _is_globalzone():
    """
    Check if we are running in the globalzone
    """
    if not __grains__["kernel"] == "SunOS":
        return False

    zonename = __salt__["cmd.run_all"]("zonename")
    if zonename["retcode"]:
        return False
    if zonename["stdout"] == "global":
        return True

    return False


def __virtual__():
    """
    We are available if we are have zonecfg and are the global zone on
    Solaris 10, OmniOS, OpenIndiana, OpenSolaris, or Smartos.
    """
    if _is_globalzone() and salt.utils.path.which("zonecfg"):
        if __grains__["os"] in ["OpenSolaris", "SmartOS", "OmniOS", "OpenIndiana"]:
            return __virtualname__
        elif (
            __grains__["os"] == "Oracle Solaris"
            and int(__grains__["osmajorrelease"]) == 10
        ):
            return __virtualname__
    return (
        False,
        "{} module can only be loaded in a solaris globalzone.".format(__virtualname__),
    )


def _clean_message(message):
    """Internal helper to sanitize message output"""
    message = message.replace("zonecfg: ", "")
    message = message.splitlines()
    for line in message:
        if line.startswith("On line"):
            message.remove(line)
    return "\n".join(message)


def _parse_value(value):
    """Internal helper for parsing configuration values into python values"""
    if isinstance(value, bool):
        return "true" if value else "false"
    elif isinstance(value, str):
        # parse compacted notation to dict
        listparser = re.compile(r"""((?:[^,"']|"[^"]*"|'[^']*')+)""")

        value = value.strip()
        if value.startswith("[") and value.endswith("]"):
            return listparser.split(value[1:-1])[1::2]
        elif value.startswith("(") and value.endswith(")"):
            rval = {}
            for pair in listparser.split(value[1:-1])[1::2]:
                pair = pair.split("=")
                if '"' in pair[1]:
                    pair[1] = pair[1].replace('"', "")
                if pair[1].isdigit():
                    rval[pair[0]] = int(pair[1])
                elif pair[1] == "true":
                    rval[pair[0]] = True
                elif pair[1] == "false":
                    rval[pair[0]] = False
                else:
                    rval[pair[0]] = pair[1]
            return rval
        else:
            if '"' in value:
                value = value.replace('"', "")
            if value.isdigit():
                return int(value)
            elif value == "true":
                return True
            elif value == "false":
                return False
            else:
                return value
    else:
        return value


def _sanitize_value(value):
    """Internal helper for converting pythonic values to configuration file values"""
    # dump dict into compated
    if isinstance(value, dict):
        new_value = []
        new_value.append("(")
        for k, v in value.items():
            new_value.append(k)
            new_value.append("=")
            new_value.append(v)
            new_value.append(",")
        new_value.append(")")
        return "".join(str(v) for v in new_value).replace(",)", ")")
    elif isinstance(value, list):
        new_value = []
        new_value.append("(")
        for item in value:
            if isinstance(item, OrderedDict):
                item = dict(item)
                for k, v in item.items():
                    new_value.append(k)
                    new_value.append("=")
                    new_value.append(v)
            else:
                new_value.append(item)
            new_value.append(",")
        new_value.append(")")
        return "".join(str(v) for v in new_value).replace(",)", ")")
    else:
        # note: we can't use shelx or pipes quote here because it makes zonecfg barf
        return '"{}"'.format(value) if " " in value else value


def _dump_cfg(cfg_file):
    """Internal helper for debugging cfg files"""
    if __salt__["file.file_exists"](cfg_file):
        with salt.utils.files.fopen(cfg_file, "r") as fp_:
            log.debug(
                "zonecfg - configuration file:\n%s",
                "".join(salt.utils.data.decode(fp_.readlines())),
            )


def create(zone, brand, zonepath, force=False):
    """
    Create an in-memory configuration for the specified zone.

    zone : string
        name of zone
    brand : string
        brand name
    zonepath : string
        path of zone
    force : boolean
        overwrite configuration

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.create deathscythe ipkg /zones/deathscythe
    """
    ret = {"status": True}

    # write config
    cfg_file = salt.utils.files.mkstemp()
    with salt.utils.files.fpopen(cfg_file, "w+", mode=0o600) as fp_:
        fp_.write("create -b -F\n" if force else "create -b\n")
        fp_.write("set brand={}\n".format(_sanitize_value(brand)))
        fp_.write("set zonepath={}\n".format(_sanitize_value(zonepath)))

    # create
    if not __salt__["file.directory_exists"](zonepath):
        __salt__["file.makedirs_perms"](
            zonepath if zonepath[-1] == "/" else "{}/".format(zonepath), mode="0700"
        )

    _dump_cfg(cfg_file)
    res = __salt__["cmd.run_all"](
        "zonecfg -z {zone} -f {cfg}".format(
            zone=zone,
            cfg=cfg_file,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    if ret["message"] == "":
        del ret["message"]
    else:
        ret["message"] = _clean_message(ret["message"])

    # cleanup config file
    if __salt__["file.file_exists"](cfg_file):
        __salt__["file.remove"](cfg_file)

    return ret


def create_from_template(zone, template):
    """
    Create an in-memory configuration from a template for the specified zone.

    zone : string
        name of zone
    template : string
        name of template

    .. warning::
        existing config will be overwritten!

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.create_from_template leo tallgeese
    """
    ret = {"status": True}

    # create from template
    _dump_cfg(template)
    res = __salt__["cmd.run_all"](
        "zonecfg -z {zone} create -t {tmpl} -F".format(
            zone=zone,
            tmpl=template,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    if ret["message"] == "":
        del ret["message"]
    else:
        ret["message"] = _clean_message(ret["message"])

    return ret


def delete(zone):
    """
    Delete the specified configuration from memory and stable storage.

    zone : string
        name of zone

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.delete epyon
    """
    ret = {"status": True}

    # delete zone
    res = __salt__["cmd.run_all"](
        "zonecfg -z {zone} delete -F".format(
            zone=zone,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    if ret["message"] == "":
        del ret["message"]
    else:
        ret["message"] = _clean_message(ret["message"])

    return ret


def export(zone, path=None):
    """
    Export the configuration from memory to stable storage.

    zone : string
        name of zone
    path : string
        path of file to export to

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.export epyon
        salt '*' zonecfg.export epyon /zones/epyon.cfg
    """
    ret = {"status": True}

    # export zone
    res = __salt__["cmd.run_all"](
        "zonecfg -z {zone} export{path}".format(
            zone=zone,
            path=" -f {}".format(path) if path else "",
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    if ret["message"] == "":
        del ret["message"]
    else:
        ret["message"] = _clean_message(ret["message"])

    return ret


def import_(zone, path):
    """
    Import the configuration to memory from stable storage.

    zone : string
        name of zone
    path : string
        path of file to export to

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.import epyon /zones/epyon.cfg
    """
    ret = {"status": True}

    # create from file
    _dump_cfg(path)
    res = __salt__["cmd.run_all"](
        "zonecfg -z {zone} -f {path}".format(
            zone=zone,
            path=path,
        )
    )
    ret["status"] = res["retcode"] == 0
    ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
    if ret["message"] == "":
        del ret["message"]
    else:
        ret["message"] = _clean_message(ret["message"])

    return ret


def _property(methode, zone, key, value):
    """
    internal handler for set and clear_property

    methode : string
        either set, add, or clear
    zone : string
        name of zone
    key : string
        name of property
    value : string
        value of property

    """
    ret = {"status": True}

    # generate update script
    cfg_file = None
    if methode not in ["set", "clear"]:
        ret["status"] = False
        ret["message"] = "unkown methode {}!".format(methode)
    else:
        cfg_file = salt.utils.files.mkstemp()
        with salt.utils.files.fpopen(cfg_file, "w+", mode=0o600) as fp_:
            if methode == "set":
                if isinstance(value, dict) or isinstance(value, list):
                    value = _sanitize_value(value)
                value = str(value).lower() if isinstance(value, bool) else str(value)
                fp_.write("{} {}={}\n".format(methode, key, _sanitize_value(value)))
            elif methode == "clear":
                fp_.write("{} {}\n".format(methode, key))

    # update property
    if cfg_file:
        _dump_cfg(cfg_file)
        res = __salt__["cmd.run_all"](
            "zonecfg -z {zone} -f {path}".format(
                zone=zone,
                path=cfg_file,
            )
        )
        ret["status"] = res["retcode"] == 0
        ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
        if ret["message"] == "":
            del ret["message"]
        else:
            ret["message"] = _clean_message(ret["message"])

        # cleanup config file
        if __salt__["file.file_exists"](cfg_file):
            __salt__["file.remove"](cfg_file)

    return ret


def set_property(zone, key, value):
    """
    Set a property

    zone : string
        name of zone
    key : string
        name of property
    value : string
        value of property

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.set_property deathscythe cpu-shares 100
    """
    return _property(
        "set",
        zone,
        key,
        value,
    )


def clear_property(zone, key):
    """
    Clear a property

    zone : string
        name of zone
    key : string
        name of property

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.clear_property deathscythe cpu-shares
    """
    return _property(
        "clear",
        zone,
        key,
        None,
    )


def _resource(methode, zone, resource_type, resource_selector, **kwargs):
    """
    internal resource hanlder

    methode : string
        add or update
    zone : string
        name of zone
    resource_type : string
        type of resource
    resource_selector : string
        unique resource identifier
    **kwargs : string|int|...
        resource properties

    """
    ret = {"status": True}

    # parse kwargs
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    for k in kwargs:
        if isinstance(kwargs[k], dict) or isinstance(kwargs[k], list):
            kwargs[k] = _sanitize_value(kwargs[k])
    if methode not in ["add", "update"]:
        ret["status"] = False
        ret["message"] = "unknown methode {}".format(methode)
        return ret
    if methode in ["update"] and resource_selector and resource_selector not in kwargs:
        ret["status"] = False
        ret["message"] = "resource selector {} not found in parameters".format(
            resource_selector
        )
        return ret

    # generate update script
    cfg_file = salt.utils.files.mkstemp()
    with salt.utils.files.fpopen(cfg_file, "w+", mode=0o600) as fp_:
        if methode in ["add"]:
            fp_.write("add {}\n".format(resource_type))
        elif methode in ["update"]:
            if resource_selector:
                value = kwargs[resource_selector]
                if isinstance(value, dict) or isinstance(value, list):
                    value = _sanitize_value(value)
                value = str(value).lower() if isinstance(value, bool) else str(value)
                fp_.write(
                    "select {} {}={}\n".format(
                        resource_type, resource_selector, _sanitize_value(value)
                    )
                )
            else:
                fp_.write("select {}\n".format(resource_type))
        for k, v in kwargs.items():
            if methode in ["update"] and k == resource_selector:
                continue
            if isinstance(v, dict) or isinstance(v, list):
                value = _sanitize_value(value)
            value = str(v).lower() if isinstance(v, bool) else str(v)
            if k in _zonecfg_resource_setters[resource_type]:
                fp_.write("set {}={}\n".format(k, _sanitize_value(value)))
            else:
                fp_.write("add {} {}\n".format(k, _sanitize_value(value)))
        fp_.write("end\n")

    # update property
    if cfg_file:
        _dump_cfg(cfg_file)
        res = __salt__["cmd.run_all"](
            "zonecfg -z {zone} -f {path}".format(
                zone=zone,
                path=cfg_file,
            )
        )
        ret["status"] = res["retcode"] == 0
        ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
        if ret["message"] == "":
            del ret["message"]
        else:
            ret["message"] = _clean_message(ret["message"])

        # cleanup config file
        if __salt__["file.file_exists"](cfg_file):
            __salt__["file.remove"](cfg_file)

    return ret


def add_resource(zone, resource_type, **kwargs):
    """
    Add a resource

    zone : string
        name of zone
    resource_type : string
        type of resource
    kwargs : string|int|...
        resource properties

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.add_resource tallgeese rctl name=zone.max-locked-memory value='(priv=privileged,limit=33554432,action=deny)'
    """
    return _resource("add", zone, resource_type, None, **kwargs)


def update_resource(zone, resource_type, resource_selector, **kwargs):
    """
    Add a resource

    zone : string
        name of zone
    resource_type : string
        type of resource
    resource_selector : string
        unique resource identifier
    kwargs : string|int|...
        resource properties

    .. note::
        Set resource_selector to None for resource that do not require one.

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.update_resource tallgeese rctl name name=zone.max-locked-memory value='(priv=privileged,limit=33554432,action=deny)'
    """
    return _resource("update", zone, resource_type, resource_selector, **kwargs)


def remove_resource(zone, resource_type, resource_key, resource_value):
    """
    Remove a resource

    zone : string
        name of zone
    resource_type : string
        type of resource
    resource_key : string
        key for resource selection
    resource_value : string
        value for resource selection

    .. note::
        Set resource_selector to None for resource that do not require one.

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.remove_resource tallgeese rctl name zone.max-locked-memory
    """
    ret = {"status": True}

    # generate update script
    cfg_file = salt.utils.files.mkstemp()
    with salt.utils.files.fpopen(cfg_file, "w+", mode=0o600) as fp_:
        if resource_key:
            fp_.write(
                "remove {} {}={}\n".format(
                    resource_type, resource_key, _sanitize_value(resource_value)
                )
            )
        else:
            fp_.write("remove {}\n".format(resource_type))

    # update property
    if cfg_file:
        _dump_cfg(cfg_file)
        res = __salt__["cmd.run_all"](
            "zonecfg -z {zone} -f {path}".format(
                zone=zone,
                path=cfg_file,
            )
        )
        ret["status"] = res["retcode"] == 0
        ret["message"] = res["stdout"] if ret["status"] else res["stderr"]
        if ret["message"] == "":
            del ret["message"]
        else:
            ret["message"] = _clean_message(ret["message"])

        # cleanup config file
        if __salt__["file.file_exists"](cfg_file):
            __salt__["file.remove"](cfg_file)

    return ret


def info(zone, show_all=False):
    """
    Display the configuration from memory

    zone : string
        name of zone
    show_all : boolean
        also include calculated values like capped-cpu, cpu-shares, ...

    CLI Example:

    .. code-block:: bash

        salt '*' zonecfg.info tallgeese
    """
    ret = {}

    # dump zone
    res = __salt__["cmd.run_all"](
        "zonecfg -z {zone} info".format(
            zone=zone,
        )
    )
    if res["retcode"] == 0:
        # parse output
        resname = None
        resdata = {}
        for line in res["stdout"].split("\n"):
            # skip some bad data
            if ":" not in line:
                continue

            # skip calculated values (if requested)
            if line.startswith("["):
                if not show_all:
                    continue
                line = line.rstrip()[1:-1]

            # extract key
            key = line.strip().split(":")[0]
            if "[" in key:
                key = key[1:]

            # parse calculated resource (if requested)
            if key in _zonecfg_info_resources_calculated:
                if resname:
                    ret[resname].append(resdata)
                if show_all:
                    resname = key
                    resdata = {}
                    if key not in ret:
                        ret[key] = []
                else:
                    resname = None
                    resdata = {}
            # parse resources
            elif key in _zonecfg_info_resources:
                if resname:
                    ret[resname].append(resdata)
                resname = key
                resdata = {}
                if key not in ret:
                    ret[key] = []
            # store resource property
            elif line.startswith("\t"):
                # ship calculated values (if requested)
                if line.strip().startswith("["):
                    if not show_all:
                        continue
                    line = line.strip()[1:-1]
                if key == "property":  # handle special 'property' keys
                    if "property" not in resdata:
                        resdata[key] = {}
                    kv = _parse_value(line.strip()[line.strip().index(":") + 1 :])
                    if "name" in kv and "value" in kv:
                        resdata[key][kv["name"]] = kv["value"]
                    else:
                        log.warning("zonecfg.info - not sure how to deal with: %s", kv)
                else:
                    resdata[key] = _parse_value(
                        line.strip()[line.strip().index(":") + 1 :]
                    )
            # store property
            else:
                if resname:
                    ret[resname].append(resdata)
                resname = None
                resdata = {}
                if key == "property":  # handle special 'property' keys
                    if "property" not in ret:
                        ret[key] = {}
                    kv = _parse_value(line.strip()[line.strip().index(":") + 1 :])
                    if "name" in kv and "value" in kv:
                        res[key][kv["name"]] = kv["value"]
                    else:
                        log.warning("zonecfg.info - not sure how to deal with: %s", kv)
                else:
                    ret[key] = _parse_value(line.strip()[line.strip().index(":") + 1 :])
        # store hanging resource
        if resname:
            ret[resname].append(resdata)

    return ret


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



"""
NAPALM NTP
==========

Manages NTP on network devices.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net> & Jerome Fleury <jf@cloudflare.com>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------
- :mod:`NAPALM proxy minion <salt.proxy.napalm>`
- :mod:`NET basic features <salt.modules.napalm_network>`

.. seealso::
    :mod:`NTP peers management state <salt.states.netntp>`

.. versionadded:: 2016.11.0
"""


import logging

# import NAPALM utils
import salt.utils.napalm
from salt.utils.napalm import proxy_napalm_wrap

log = logging.getLogger(__file__)


# ----------------------------------------------------------------------------------------------------------------------
# module properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = "ntp"
__proxyenabled__ = ["napalm"]
__virtual_aliases__ = ("napalm_ntp",)
# uses NAPALM-based proxy to interact with network devices

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    """
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


@proxy_napalm_wrap
def peers(**kwargs):  # pylint: disable=unused-argument

    """
    Returns a list the NTP peers configured on the network device.

    :return: configured NTP peers as list.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.peers

    Example output:

    .. code-block:: python

        [
            '192.168.0.1',
            '172.17.17.1',
            '172.17.17.2',
            '2400:cb00:6:1024::c71b:840a'
        ]

    """

    ntp_peers = salt.utils.napalm.call(
        napalm_device, "get_ntp_peers", **{}  # pylint: disable=undefined-variable
    )

    if not ntp_peers.get("result"):
        return ntp_peers

    ntp_peers_list = list(ntp_peers.get("out", {}).keys())

    ntp_peers["out"] = ntp_peers_list

    return ntp_peers


@proxy_napalm_wrap
def servers(**kwargs):  # pylint: disable=unused-argument

    """
    Returns a list of the configured NTP servers on the device.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.servers

    Example output:

    .. code-block:: python

        [
            '192.168.0.1',
            '172.17.17.1',
            '172.17.17.2',
            '2400:cb00:6:1024::c71b:840a'
        ]
    """

    ntp_servers = salt.utils.napalm.call(
        napalm_device, "get_ntp_servers", **{}  # pylint: disable=undefined-variable
    )

    if not ntp_servers.get("result"):
        return ntp_servers

    ntp_servers_list = list(ntp_servers.get("out", {}).keys())

    ntp_servers["out"] = ntp_servers_list

    return ntp_servers


@proxy_napalm_wrap
def stats(peer=None, **kwargs):  # pylint: disable=unused-argument

    """
    Returns a dictionary containing synchronization details of the NTP peers.

    :param peer: Returns only the details of a specific NTP peer.
    :return: a list of dictionaries, with the following keys:

        * remote
        * referenceid
        * synchronized
        * stratum
        * type
        * when
        * hostpoll
        * reachability
        * delay
        * offset
        * jitter

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.stats

    Example output:

    .. code-block:: python

        [
            {
                'remote'        : '188.114.101.4',
                'referenceid'   : '188.114.100.1',
                'synchronized'  : True,
                'stratum'       : 4,
                'type'          : '-',
                'when'          : '107',
                'hostpoll'      : 256,
                'reachability'  : 377,
                'delay'         : 164.228,
                'offset'        : -13.866,
                'jitter'        : 2.695
            }
        ]
    """

    proxy_output = salt.utils.napalm.call(
        napalm_device, "get_ntp_stats", **{}  # pylint: disable=undefined-variable
    )

    if not proxy_output.get("result"):
        return proxy_output

    ntp_peers = proxy_output.get("out")

    if peer:
        ntp_peers = [
            ntp_peer for ntp_peer in ntp_peers if ntp_peer.get("remote", "") == peer
        ]

    proxy_output.update({"out": ntp_peers})

    return proxy_output


@proxy_napalm_wrap
def set_peers(*peers, **options):

    """
    Configures a list of NTP peers on the device.

    :param peers: list of IP Addresses/Domain Names
    :param test (bool): discard loaded config. By default ``test`` is False
        (will not dicard the changes)
    :commit commit (bool): commit loaded config. By default ``commit`` is True
        (will commit the changes). Useful when the user does not want to commit
        after each change, but after a couple.

    By default this function will commit the config changes (if any). To load without committing, use the `commit`
    option. For dry run use the `test` argument.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.set_peers 192.168.0.1 172.17.17.1 time.apple.com
        salt '*' ntp.set_peers 172.17.17.1 test=True  # only displays the diff
        salt '*' ntp.set_peers 192.168.0.1 commit=False  # preserves the changes, but does not commit
    """

    test = options.pop("test", False)
    commit = options.pop("commit", True)

    # pylint: disable=undefined-variable
    return __salt__["net.load_template"](
        "set_ntp_peers",
        peers=peers,
        test=test,
        commit=commit,
        inherit_napalm_device=napalm_device,
    )
    # pylint: enable=undefined-variable


@proxy_napalm_wrap
def set_servers(*servers, **options):
    """
    Configures a list of NTP servers on the device.

    :param servers: list of IP Addresses/Domain Names
    :param test (bool): discard loaded config. By default ``test`` is False
        (will not dicard the changes)
    :commit commit (bool): commit loaded config. By default ``commit`` is True
        (will commit the changes). Useful when the user does not want to commit
        after each change, but after a couple.

    By default this function will commit the config changes (if any). To load without committing, use the `commit`
    option. For dry run use the `test` argument.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.set_servers 192.168.0.1 172.17.17.1 time.apple.com
        salt '*' ntp.set_servers 172.17.17.1 test=True  # only displays the diff
        salt '*' ntp.set_servers 192.168.0.1 commit=False  # preserves the changes, but does not commit
    """

    test = options.pop("test", False)
    commit = options.pop("commit", True)

    # pylint: disable=undefined-variable
    return __salt__["net.load_template"](
        "set_ntp_servers",
        servers=servers,
        test=test,
        commit=commit,
        inherit_napalm_device=napalm_device,
    )
    # pylint: enable=undefined-variable


@proxy_napalm_wrap
def delete_peers(*peers, **options):

    """
    Removes NTP peers configured on the device.

    :param peers: list of IP Addresses/Domain Names to be removed as NTP peers
    :param test (bool): discard loaded config. By default ``test`` is False
        (will not dicard the changes)
    :param commit (bool): commit loaded config. By default ``commit`` is True
        (will commit the changes). Useful when the user does not want to commit
        after each change, but after a couple.

    By default this function will commit the config changes (if any). To load
    without committing, use the ``commit`` option. For a dry run, use the
    ``test`` argument.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.delete_peers 8.8.8.8 time.apple.com
        salt '*' ntp.delete_peers 172.17.17.1 test=True  # only displays the diff
        salt '*' ntp.delete_peers 192.168.0.1 commit=False  # preserves the changes, but does not commit
    """

    test = options.pop("test", False)
    commit = options.pop("commit", True)

    # pylint: disable=undefined-variable
    return __salt__["net.load_template"](
        "delete_ntp_peers",
        peers=peers,
        test=test,
        commit=commit,
        inherit_napalm_device=napalm_device,
    )
    # pylint: enable=undefined-variable


@proxy_napalm_wrap
def delete_servers(*servers, **options):

    """
    Removes NTP servers configured on the device.

    :param servers: list of IP Addresses/Domain Names to be removed as NTP
        servers
    :param test (bool): discard loaded config. By default ``test`` is False
        (will not dicard the changes)
    :param commit (bool): commit loaded config. By default ``commit`` is True
        (will commit the changes). Useful when the user does not want to commit
        after each change, but after a couple.

    By default this function will commit the config changes (if any). To load
    without committing, use the ``commit`` option. For dry run use the ``test``
    argument.

    CLI Example:

    .. code-block:: bash

        salt '*' ntp.delete_servers 8.8.8.8 time.apple.com
        salt '*' ntp.delete_servers 172.17.17.1 test=True  # only displays the diff
        salt '*' ntp.delete_servers 192.168.0.1 commit=False  # preserves the changes, but does not commit
    """

    test = options.pop("test", False)
    commit = options.pop("commit", True)

    # pylint: disable=undefined-variable
    return __salt__["net.load_template"](
        "delete_ntp_servers",
        servers=servers,
        test=test,
        commit=commit,
        inherit_napalm_device=napalm_device,
    )
    # pylint: enable=undefined-variable



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.cpan as cpan
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class CpanTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.cpan
    """

    # 'install' function tests: 2

    def setup_loader_modules(self):
        return {cpan: {}}

    def test_install(self):
        """
        Test if it install a module from cpan
        """
        mock = MagicMock(return_value="")
        with patch.dict(cpan.__salt__, {"cmd.run": mock}):
            mock = MagicMock(
                side_effect=[{"installed version": None}, {"installed version": "3.1"}]
            )
            with patch.object(cpan, "show", mock):
                self.assertDictEqual(cpan.install("Alloy"), {"new": "3.1", "old": None})

    def test_install_error(self):
        """
        Test if it install a module from cpan
        """
        mock = MagicMock(return_value="don't know what it is")
        with patch.dict(cpan.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(
                cpan.install("Alloy"),
                {
                    "error": "CPAN cannot identify this package",
                    "new": None,
                    "old": None,
                },
            )

    # 'remove' function tests: 4

    def test_remove(self):
        """
        Test if it remove a module using cpan
        """
        with patch("os.listdir", MagicMock(return_value=[""])):
            mock = MagicMock(return_value="")
            with patch.dict(cpan.__salt__, {"cmd.run": mock}):
                mock = MagicMock(
                    return_value={
                        "installed version": "2.1",
                        "cpan build dirs": [""],
                        "installed file": "/root",
                    }
                )
                with patch.object(cpan, "show", mock):
                    self.assertDictEqual(
                        cpan.remove("Alloy"), {"new": None, "old": "2.1"}
                    )

    def test_remove_unexist_error(self):
        """
        Test if it try to remove an unexist module using cpan
        """
        mock = MagicMock(return_value="don't know what it is")
        with patch.dict(cpan.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(
                cpan.remove("Alloy"), {"error": "This package does not seem to exist"}
            )

    def test_remove_noninstalled_error(self):
        """
        Test if it remove non installed module using cpan
        """
        mock = MagicMock(return_value={"installed version": None})
        with patch.object(cpan, "show", mock):
            self.assertDictEqual(cpan.remove("Alloy"), {"new": None, "old": None})

    def test_remove_nopan_error(self):
        """
        Test if it gives no cpan error while removing
        """
        ret = {"error": "No CPAN data available to use for uninstalling"}
        mock = MagicMock(return_value={"installed version": "2.1"})
        with patch.object(cpan, "show", mock):
            self.assertDictEqual(cpan.remove("Alloy"), ret)

    # 'list' function tests: 1

    def test_list(self):
        """
        Test if it list installed Perl module
        """
        mock = MagicMock(return_value="")
        with patch.dict(cpan.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(cpan.list_(), {})

    # 'show' function tests: 2

    def test_show(self):
        """
        Test if it show information about a specific Perl module
        """
        mock = MagicMock(return_value="")
        with patch.dict(cpan.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(
                cpan.show("Alloy"),
                {"error": "This package does not seem to exist", "name": "Alloy"},
            )

    def test_show_mock(self):
        """
        Test if it show information about a specific Perl module
        """
        with patch("salt.modules.cpan.show", MagicMock(return_value={"Salt": "salt"})):
            mock = MagicMock(return_value="Salt module installed")
            with patch.dict(cpan.__salt__, {"cmd.run": mock}):
                self.assertDictEqual(cpan.show("Alloy"), {"Salt": "salt"})

    # 'show_config' function tests: 1

    def test_show_config(self):
        """
        Test if it return a dict of CPAN configuration values
        """
        mock = MagicMock(return_value="")
        with patch.dict(cpan.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(cpan.show_config(), {})



"""
    :codeauthor: Alexander Pyatkin <asp@thexyz.net>
"""


import pytest
import salt.states.webutil as htpasswd
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {htpasswd: {"__opts__": {"test": False}}}


def test_user_exists_already():
    """
    Test if it returns True when user already exists in htpasswd file
    """

    mock = MagicMock(return_value={"retcode": 0})

    with patch.dict(htpasswd.__salt__, {"file.grep": mock}):
        ret = htpasswd.user_exists("larry", "badpass", "/etc/httpd/htpasswd")
        expected = {
            "name": "larry",
            "result": True,
            "comment": "User already known",
            "changes": {},
        }
        assert ret == expected


def test_new_user_success():
    """
    Test if it returns True when new user is added to htpasswd file
    """

    mock_grep = MagicMock(return_value={"retcode": 1})
    mock_useradd = MagicMock(return_value={"retcode": 0, "stderr": "Success"})

    with patch.dict(
        htpasswd.__salt__, {"file.grep": mock_grep, "webutil.useradd": mock_useradd}
    ):
        ret = htpasswd.user_exists("larry", "badpass", "/etc/httpd/htpasswd")
        expected = {
            "name": "larry",
            "result": True,
            "comment": "Success",
            "changes": {"larry": True},
        }
        assert ret == expected


def test_new_user_error():
    """
    Test if it returns False when adding user to htpasswd failed
    """

    mock_grep = MagicMock(return_value={"retcode": 1})
    mock_useradd = MagicMock(return_value={"retcode": 1, "stderr": "Error"})

    with patch.dict(
        htpasswd.__salt__, {"file.grep": mock_grep, "webutil.useradd": mock_useradd}
    ):
        ret = htpasswd.user_exists("larry", "badpass", "/etc/httpd/htpasswd")
        expected = {
            "name": "larry",
            "result": False,
            "comment": "Error",
            "changes": {},
        }
        assert ret == expected



"""
tests.unit.returners.test_highstate_return
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the Highstate Returner Cache.
"""


import json
import logging
import os

import salt.returners.highstate_return as highstate
import salt.utils.files
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class HighstateReturnerTestCase(TestCase, LoaderModuleMockMixin):
    """
    Tests for the highstate_return returner
    """

    output_file = os.path.join(RUNTIME_VARS.TMP, "highstate_return")

    def tearDown(self):
        os.unlink(self.output_file)

    def setup_loader_modules(self):
        return {
            highstate: {
                "__opts__": {
                    "highstate.report_everything": True,
                    "highstate.report_format": "json",
                    "highstate.report_delivery": "file",
                    "highstate.file_output": self.output_file,
                }
            }
        }

    def test_pipe_in_name(self):
        ret = {
            "fun_args": ["test"],
            "jid": "20180308201402941603",
            "return": {
                "cmd_|-test_|-echo hi | grep h\n_|-run": {
                    "comment": 'Command "echo hi | grep h\n" run',
                    "name": "echo hi | grep h\n",
                    "start_time": "20:14:03.053612",
                    "result": True,
                    "duration": 75.198,
                    "__run_num__": 0,
                    "__sls__": "test",
                    "changes": {
                        "pid": 1429,
                        "retcode": 0,
                        "stderr": "",
                        "stdout": "hi",
                    },
                    "__id__": "test",
                }
            },
            "retcode": 0,
            "success": True,
            "fun": "state.apply",
            "id": "salt",
            "out": "highstate",
        }
        expected = [
            {
                "stats": [
                    {"total": 1},
                    {"failed": 0, "__style__": "failed"},
                    {"unchanged": 0, "__style__": "unchanged"},
                    {"changed": 1, "__style__": "changed"},
                    {"duration": 75.198},
                ],
            },
            {
                "job": [
                    {"function": "state.apply"},
                    {"arguments": ["test"]},
                    {"jid": "20180308201402941603"},
                    {"success": True},
                    {"retcode": 0},
                ],
            },
            {
                "states": [
                    {
                        "test": [
                            {"function": "cmd.run"},
                            {"name": "echo hi | grep h\n"},
                            {"result": True},
                            {"duration": 75.198},
                            {"comment": 'Command "echo hi | grep h\n" run'},
                            {
                                "changes": [
                                    {"pid": 1429},
                                    {"retcode": 0},
                                    {"stderr": ""},
                                    {"stdout": "hi"},
                                ]
                            },
                            {"started": "20:14:03.053612"},
                        ],
                        "__style__": "changed",
                    }
                ]
            },
        ]
        highstate.returner(ret)
        with salt.utils.files.fopen(self.output_file) as fh_:
            self.assertEqual(json.load(fh_), expected)



"""
Manage Route53 records

.. versionadded:: 2014.7.0

Create and delete Route53 records. Be aware that this interacts with Amazon's
services, and so may incur charges.

This module uses ``boto``, which can be installed via package, or pip.

This module accepts explicit route53 credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    route53.keyid: GKTADJGHEIQSXMKKRBJ08H
    route53.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile, either
passed in as a dict, or as a string to pull from pillars or minion config:

.. code-block:: yaml

    myprofile:
      keyid: GKTADJGHEIQSXMKKRBJ08H
      key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
      region: us-east-1

.. code-block:: yaml

    mycnamerecord:
      boto_route53.present:
        - name: test.example.com.
        - value: my-elb.us-east-1.elb.amazonaws.com.
        - zone: example.com.
        - ttl: 60
        - record_type: CNAME
        - region: us-east-1
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    # Using a profile from pillars
    myarecord:
      boto_route53.present:
        - name: test.example.com.
        - value: 1.1.1.1
        - zone: example.com.
        - ttl: 60
        - record_type: A
        - region: us-east-1
        - profile: myprofile

    # Passing in a profile
    myarecord:
      boto_route53.present:
        - name: test.example.com.
        - value: 1.1.1.1
        - zone: example.com.
        - ttl: 60
        - record_type: A
        - region: us-east-1
        - profile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
"""


import logging
import uuid

import salt.utils.data
import salt.utils.json
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto is available.
    """
    if "boto_route53.get_record" in __salt__:
        return "boto_route53"
    return (False, "boto_route53 module could not be loaded")


def rr_present(*args, **kwargs):
    return present(*args, **kwargs)


def present(
    name,
    value,
    zone,
    record_type,
    ttl=None,
    identifier=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    wait_for_sync=True,
    split_dns=False,
    private_zone=False,
):
    """
    Ensure the Route53 record is present.

    name
        Name of the record.

    value
        Value of the record.  As a special case, you can pass in:
            `private:<Name tag>` to have the function autodetermine the private IP
            `public:<Name tag>` to have the function autodetermine the public IP

    zone
        The zone to create the record in.

    record_type
        The record type (A, NS, MX, TXT, etc.)

    ttl
        The time to live for the record.

    identifier
        The unique identifier to use for this record.

    region
        The region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that contains a dict
        with region, key and keyid.

    wait_for_sync
        Wait for an INSYNC change status from Route53 before returning success.

    split_dns
        Route53 supports parallel public and private DNS zones with the same name.

    private_zone
        If using split_dns, specify if this is the private zone.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    # If a list is passed in for value, change it to a comma-separated string
    # So it will work with subsequent boto module calls and string functions
    if isinstance(value, list):
        value = ",".join(value)
    elif value.startswith("private:") or value.startswith("public:"):
        name_tag = value.split(":", 1)[1]
        in_states = ("pending", "rebooting", "running", "stopping", "stopped")
        r = __salt__["boto_ec2.find_instances"](
            name=name_tag, return_objs=True, in_states=in_states, profile=profile
        )
        if len(r) < 1:
            ret["comment"] = "Error: instance with Name tag {} not found".format(
                name_tag
            )
            ret["result"] = False
            return ret
        if len(r) > 1:
            ret["comment"] = "Error: Name tag {} matched more than one instance".format(
                name_tag
            )
            ret["result"] = False
            return ret
        instance = r[0]
        private_ip = getattr(instance, "private_ip_address", None)
        public_ip = getattr(instance, "ip_address", None)
        if value.startswith("private:"):
            value = private_ip
            log.info("Found private IP %s for instance %s", private_ip, name_tag)
        else:
            if public_ip is None:
                ret[
                    "comment"
                ] = "Error: No Public IP assigned to instance with Name {}".format(
                    name_tag
                )
                ret["result"] = False
                return ret
            value = public_ip
            log.info("Found public IP %s for instance %s", public_ip, name_tag)

    try:
        record = __salt__["boto_route53.get_record"](
            name,
            zone,
            record_type,
            False,
            region,
            key,
            keyid,
            profile,
            split_dns,
            private_zone,
            identifier,
        )
    except SaltInvocationError as err:
        ret["comment"] = "Error: {}".format(err)
        ret["result"] = False
        return ret

    if isinstance(record, dict) and not record:
        if __opts__["test"]:
            ret["comment"] = "Route53 record {} set to be added.".format(name)
            ret["result"] = None
            return ret
        added = __salt__["boto_route53.add_record"](
            name,
            value,
            zone,
            record_type,
            identifier,
            ttl,
            region,
            key,
            keyid,
            profile,
            wait_for_sync,
            split_dns,
            private_zone,
        )
        if added:
            ret["changes"]["old"] = None
            ret["changes"]["new"] = {
                "name": name,
                "value": value,
                "record_type": record_type,
                "ttl": ttl,
                "identifier": identifier,
            }
            ret["comment"] = "Added {} Route53 record.".format(name)
        else:
            ret["result"] = False
            ret["comment"] = "Failed to add {} Route53 record.".format(name)
            return ret
    elif record:
        need_to_update = False
        # Values can be a comma separated list and some values will end with a
        # period (even if we set it without one). To easily check this we need
        # to split and check with the period stripped from the input and what's
        # in route53.
        # TODO: figure out if this will cause us problems with some records.
        _values = [x.rstrip(".") for x in value.split(",")]
        _r_values = [x.rstrip(".") for x in record["value"].split(",")]
        _values.sort()
        _r_values.sort()
        if _values != _r_values:
            need_to_update = True
        if identifier and identifier != record["identifier"]:
            need_to_update = True
        if ttl and str(ttl) != str(record["ttl"]):
            need_to_update = True
        if need_to_update:
            if __opts__["test"]:
                ret["comment"] = "Route53 record {} set to be updated.".format(name)
                ret["result"] = None
                return ret
            updated = __salt__["boto_route53.update_record"](
                name,
                value,
                zone,
                record_type,
                identifier,
                ttl,
                region,
                key,
                keyid,
                profile,
                wait_for_sync,
                split_dns,
                private_zone,
            )
            if updated:
                ret["changes"]["old"] = record
                ret["changes"]["new"] = {
                    "name": name,
                    "value": value,
                    "record_type": record_type,
                    "ttl": ttl,
                    "identifier": identifier,
                }
                ret["comment"] = "Updated {} Route53 record.".format(name)
            else:
                ret["result"] = False
                ret["comment"] = "Failed to update {} Route53 record.".format(name)
        else:
            ret["comment"] = "{} exists.".format(name)
    return ret


def rr_absent(*args, **kwargs):
    return absent(*args, **kwargs)


def absent(
    name,
    zone,
    record_type,
    identifier=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    wait_for_sync=True,
    split_dns=False,
    private_zone=False,
):
    """
    Ensure the Route53 record is deleted.

    name
        Name of the record.

    zone
        The zone to delete the record from.

    record_type
        The record type (A, NS, MX, TXT, etc.)

    identifier
        An identifier to match for deletion.

    region
        The region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.

    wait_for_sync
        Wait for an INSYNC change status from Route53.

    split_dns
        Route53 supports a public and private DNS zone with the same
        names.

    private_zone
        If using split_dns, specify if this is the private zone.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    record = __salt__["boto_route53.get_record"](
        name,
        zone,
        record_type,
        False,
        region,
        key,
        keyid,
        profile,
        split_dns,
        private_zone,
        identifier,
    )
    if record:
        if __opts__["test"]:
            ret["comment"] = "Route53 record {} set to be deleted.".format(name)
            ret["result"] = None
            return ret
        deleted = __salt__["boto_route53.delete_record"](
            name,
            zone,
            record_type,
            identifier,
            False,
            region,
            key,
            keyid,
            profile,
            wait_for_sync,
            split_dns,
            private_zone,
        )
        if deleted:
            ret["changes"]["old"] = record
            ret["changes"]["new"] = None
            ret["comment"] = "Deleted {} Route53 record.".format(name)
        else:
            ret["result"] = False
            ret["comment"] = "Failed to delete {} Route53 record.".format(name)
    else:
        ret["comment"] = "{} does not exist.".format(name)
    return ret


def hosted_zone_present(
    name,
    domain_name=None,
    private_zone=False,
    caller_ref=None,
    comment="",
    vpc_id=None,
    vpc_name=None,
    vpc_region=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure a hosted zone exists with the given attributes. Note that most
    things cannot be modified once a zone is created - it must be deleted and
    re-spun to update these attributes:

    - private_zone (AWS API limitation).
    - comment (the appropriate call exists in the AWS API and in boto3, but has
      not, as of this writing, been added to boto2).
    - vpc_id (same story - we really need to rewrite this module with boto3)
    - vpc_name (really just a pointer to vpc_id anyway).
    - vpc_region (again, supported in boto3 but not boto2).

    If you need the ability to update these attributes, please use the newer
    boto3_route53 module instead.

    name
        The name of the state definition.

    domain_name
        The name of the domain. This must be fully-qualified, terminating with a period.  This is
        the name you have registered with your domain registrar.  It is also the name you will
        delegate from your registrar to the Amazon Route 53 delegation servers returned in response
        to this request.  Defaults to the value of name if not provided.

    private_zone
        Set True if creating a private hosted zone.

    caller_ref
        A unique string that identifies the request and that allows create_hosted_zone() calls to be
        retried without the risk of executing the operation twice.  This helps ensure idempotency
        across state calls, but can cause issues if a zone is deleted and then an attempt is made
        to recreate it with the same caller_ref.  If not provided, a unique UUID will be generated
        at each state run, which avoids the risk of the above (transient) error.  This option is
        generally not needed.  Maximum length of 128.

    comment
        Any comments you want to include about the hosted zone.

    vpc_id
        When creating a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with vpe_name.  Ignored when creating a non-private zone.

    vpc_name
        When creating a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with vpe_id.  Ignored when creating a non-private zone.

    vpc_region
        When creating a private hosted zone, the region of the associated VPC is required.  If not
        provided, an effort will be made to determine it from vpc_id or vpc_name, where possible.
        If this fails, you'll need to provide an explicit value for this option.  Ignored when
        creating a non-private zone.
    """
    domain_name = domain_name if domain_name else name

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    # First translaste vpc_name into a vpc_id if possible
    if private_zone:
        if not salt.utils.data.exactly_one((vpc_name, vpc_id)):
            raise SaltInvocationError(
                "Either vpc_name or vpc_id is required when creating a private zone."
            )
        vpcs = __salt__["boto_vpc.describe_vpcs"](
            vpc_id=vpc_id,
            name=vpc_name,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        ).get("vpcs", [])
        if vpc_region and vpcs:
            vpcs = [v for v in vpcs if v["region"] == vpc_region]
        if not vpcs:
            msg = "Private zone requested but a VPC matching given criteria not found."
            log.error(msg)
            ret["comment"] = msg
            ret["result"] = False
            return ret
        if len(vpcs) > 1:
            log.error(
                "Private zone requested but multiple VPCs matching given "
                "criteria found: %s",
                [v["id"] for v in vpcs],
            )
            return None
        vpc = vpcs[0]
        if vpc_name:
            vpc_id = vpc["id"]
        if not vpc_region:
            vpc_region = vpc["region"]

    # Next, see if it (or they) exist at all, anywhere?
    deets = __salt__["boto_route53.describe_hosted_zones"](
        domain_name=domain_name, region=region, key=key, keyid=keyid, profile=profile
    )

    create = False
    if not deets:
        create = True
    else:  # Something exists - now does it match our criteria?
        if (
            salt.utils.json.loads(deets["HostedZone"]["Config"]["PrivateZone"])
            != private_zone
        ):
            create = True
        else:
            if private_zone:
                for v, d in deets.get("VPCs", {}).items():
                    if d["VPCId"] == vpc_id and d["VPCRegion"] == vpc_region:
                        create = False
                        break
                    else:
                        create = True
        if not create:
            ret["comment"] = "Hostd Zone {} already in desired state".format(
                domain_name
            )
        else:
            # Until we get modifies in place with boto3, the best option is to
            # attempt creation and let route53 tell us if we're stepping on
            # toes.  We can't just fail, because some scenarios (think split
            # horizon DNS) require zones with identical names but different
            # settings...
            log.info(
                "A Hosted Zone with name %s already exists, but with "
                "different settings.  Will attempt to create the one "
                "requested on the assumption this is what is desired.  "
                "This may fail...",
                domain_name,
            )

    if create:
        if caller_ref is None:
            caller_ref = str(uuid.uuid4())
        if __opts__["test"]:
            ret["comment"] = "Route53 Hosted Zone {} set to be added.".format(
                domain_name
            )
            ret["result"] = None
            return ret
        res = __salt__["boto_route53.create_hosted_zone"](
            domain_name=domain_name,
            caller_ref=caller_ref,
            comment=comment,
            private_zone=private_zone,
            vpc_id=vpc_id,
            vpc_region=vpc_region,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
        if res:
            msg = "Hosted Zone {} successfully created".format(domain_name)
            log.info(msg)
            ret["comment"] = msg
            ret["changes"]["old"] = None
            ret["changes"]["new"] = res
        else:
            ret["comment"] = "Creating Hosted Zone {} failed".format(domain_name)
            ret["result"] = False

    return ret


def hosted_zone_absent(
    name, domain_name=None, region=None, key=None, keyid=None, profile=None
):
    """
    Ensure the Route53 Hostes Zone described is absent

    name
        The name of the state definition.

    domain_name
        The FQDN (including final period) of the zone you wish absent.  If not
        provided, the value of name will be used.

    """
    domain_name = domain_name if domain_name else name

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    deets = __salt__["boto_route53.describe_hosted_zones"](
        domain_name=domain_name, region=region, key=key, keyid=keyid, profile=profile
    )
    if not deets:
        ret["comment"] = "Hosted Zone {} already absent".format(domain_name)
        log.info(ret["comment"])
        return ret
    if __opts__["test"]:
        ret["comment"] = "Route53 Hosted Zone {} set to be deleted.".format(domain_name)
        ret["result"] = None
        return ret
    # Not entirely comfortable with this - no safety checks around pub/priv, VPCs
    # or anything else.  But this is all the module function exposes, so hmph.
    # Inclined to put it on the "wait 'til we port to boto3" pile in any case :)
    if __salt__["boto_route53.delete_zone"](
        zone=domain_name, region=region, key=key, keyid=keyid, profile=profile
    ):
        ret["comment"] = "Route53 Hosted Zone {} deleted".format(domain_name)
        log.info(ret["comment"])
        ret["changes"]["old"] = deets
        ret["changes"]["new"] = None

    return ret



"""
    :codeauthor: Erik Johnson <erik@saltstack.com>
"""


import logging
import os

import salt.states.git as git_state  # Don't potentially shadow GitPython
from tests.support.helpers import with_tempdir
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import DEFAULT, MagicMock, Mock, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class GitTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.git
    """

    def setup_loader_modules(self):
        return {
            git_state: {"__env__": "base", "__opts__": {"test": False}, "__salt__": {}}
        }

    @with_tempdir()
    def test_latest_no_diff_for_bare_repo(self, target):
        """
        This test ensures that we don't attempt to diff when cloning a repo
        using either bare=True or mirror=True.
        """
        name = "https://foo.com/bar/baz.git"
        gitdir = os.path.join(target, "refs")
        isdir_mock = MagicMock(
            side_effect=lambda path: DEFAULT if path != gitdir else True
        )

        branches = ["foo", "bar", "baz"]
        tags = ["v1.1.0", "v.1.1.1", "v1.2.0"]
        local_head = "b9ef06ab6b7524eb7c27d740dbbd5109c6d75ee4"
        remote_head = "eef672c1ec9b8e613905dbcd22a4612e31162807"

        git_diff = Mock()
        dunder_salt = {
            "git.current_branch": MagicMock(return_value=branches[0]),
            "git.config_get_regexp": MagicMock(return_value={}),
            "git.diff": git_diff,
            "git.fetch": MagicMock(return_value={}),
            "git.is_worktree": MagicMock(return_value=False),
            "git.list_branches": MagicMock(return_value=branches),
            "git.list_tags": MagicMock(return_value=tags),
            "git.remote_refs": MagicMock(return_value={"HEAD": remote_head}),
            "git.remotes": MagicMock(
                return_value={"origin": {"fetch": name, "push": name}}
            ),
            "git.rev_parse": MagicMock(side_effect=git_state.CommandExecutionError()),
            "git.revision": MagicMock(return_value=local_head),
            "git.version": MagicMock(return_value="1.8.3.1"),
        }
        with patch("os.path.isdir", isdir_mock), patch.dict(
            git_state.__salt__, dunder_salt
        ):
            result = git_state.latest(
                name=name,
                target=target,
                mirror=True,  # mirror=True implies bare=True
            )
            assert result["result"] is True, result
            git_diff.assert_not_called()

    def test_latest_without_target(self):
        """
        Test latest when called without passing target
        """
        name = "https://foo.com/bar/baz.git"
        self.assertRaises(TypeError, git_state.latest, name)

    def test_detached_without_target(self):
        """
        Test detached when called without passing target
        """
        name = "https://foo.com/bar/baz.git"
        self.assertRaises(TypeError, git_state.detached, name)

    def test_cloned_without_target(self):
        """
        Test cloned when called without passing target
        """
        name = "https://foo.com/bar/baz.git"
        self.assertRaises(TypeError, git_state.cloned, name)



"""
Tests for the Azure Blob External Pillar.
"""

import os
import pickle
import tempfile
import time

import salt.config
import salt.loader
import salt.pillar.azureblob as azureblob
import salt.utils.files
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

HAS_LIBS = False
try:
    # pylint: disable=no-name-in-module
    from azure.storage.blob import BlobServiceClient

    # pylint: enable=no-name-in-module

    HAS_LIBS = True
except ImportError:
    pass


class MockBlob(dict):
    """
    Creates a Mock Blob object.
    """

    name = ""

    def __init__(self):
        super().__init__(
            {
                "container": None,
                "name": "test.sls",
                "prefix": None,
                "delimiter": "/",
                "results_per_page": None,
                "location_mode": None,
            }
        )


class MockContainerClient:
    """
    Creates a Mock ContainerClient.
    """

    def __init__(self):
        pass

    def walk_blobs(self, *args, **kwargs):
        yield MockBlob()

    def get_blob_client(self, *args, **kwargs):
        pass


class MockBlobServiceClient:
    """
    Creates a Mock BlobServiceClient.
    """

    def __init__(self):
        pass

    def get_container_client(self, *args, **kwargs):
        container_client = MockContainerClient()
        return container_client


@skipIf(HAS_LIBS is False, "The azure.storage.blob module must be installed.")
class AzureBlobTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.pillar.azureblob ext_pillar.
    """

    def setup_loader_modules(self):
        self.opts = salt.config.DEFAULT_MASTER_OPTS.copy()
        utils = salt.loader.utils(self.opts)
        return {
            azureblob: {"__opts__": self.opts, "__utils__": utils},
        }

    def test__init_expired(self):
        """
        Tests the result of _init when the cache is expired.
        """
        container = "test"
        multiple_env = False
        environment = "base"
        blob_cache_expire = 0  # The cache will be expired
        blob_client = MockBlobServiceClient()
        cache_file = tempfile.NamedTemporaryFile()
        # Patches the _get_containers_cache_filename module so that it returns the name of the new tempfile that
        # represents the cache file
        with patch.object(
            azureblob,
            "_get_containers_cache_filename",
            MagicMock(return_value=str(cache_file.name)),
        ):
            # Patches the from_connection_string module of the BlobServiceClient class so that a connection string does
            # not need to be given. Additionally it returns example blob data used by the ext_pillar.
            with patch.object(
                BlobServiceClient,
                "from_connection_string",
                MagicMock(return_value=blob_client),
            ):
                ret = azureblob._init(
                    "", container, multiple_env, environment, blob_cache_expire
                )
        cache_file.close()
        self.assertEqual(
            ret,
            {
                "base": {
                    "test": [
                        {
                            "container": None,
                            "name": "test.sls",
                            "prefix": None,
                            "delimiter": "/",
                            "results_per_page": None,
                            "location_mode": None,
                        }
                    ]
                }
            },
        )

    def test__init_not_expired(self):
        """
        Tests the result of _init when the cache is not expired.
        """
        container = "test"
        multiple_env = False
        environment = "base"
        blob_cache_expire = (time.time()) * (
            time.time()
        )  # The cache will not be expired
        metadata = {
            "base": {
                "test": [
                    {"name": "base/secret.sls", "relevant": "include.sls"},
                    {"name": "blobtest.sls", "irrelevant": "ignore.sls"},
                ]
            }
        }
        cache_file = tempfile.NamedTemporaryFile()
        # Pickles the metadata and stores it in cache_file
        with salt.utils.files.fopen(str(cache_file), "wb") as fp_:
            pickle.dump(metadata, fp_)
        # Patches the _get_containers_cache_filename module so that it returns the name of the new tempfile that
        # represents the cache file
        with patch.object(
            azureblob,
            "_get_containers_cache_filename",
            MagicMock(return_value=str(cache_file.name)),
        ):
            # Patches the _read_containers_cache_file module so that it returns what it normally would if the new
            # tempfile representing the cache file was passed to it
            plugged = azureblob._read_containers_cache_file(str(cache_file))
            with patch.object(
                azureblob,
                "_read_containers_cache_file",
                MagicMock(return_value=plugged),
            ):
                ret = azureblob._init(
                    "", container, multiple_env, environment, blob_cache_expire
                )
        fp_.close()
        os.remove(str(fp_.name))
        cache_file.close()
        self.assertEqual(ret, metadata)

    def test__get_cache_dir(self):
        """
        Tests the result of _get_cache_dir.
        """
        ret = azureblob._get_cache_dir()
        self.assertEqual(ret, "/var/cache/salt/master/pillar_azureblob")

    def test__get_cached_file_name(self):
        """
        Tests the result of _get_cached_file_name.
        """
        container = "test"
        saltenv = "base"
        path = "base/secret.sls"
        ret = azureblob._get_cached_file_name(container, saltenv, path)
        self.assertEqual(
            ret, "/var/cache/salt/master/pillar_azureblob/base/test/base/secret.sls"
        )

    def test__get_containers_cache_filename(self):
        """
        Tests the result of _get_containers_cache_filename.
        """
        container = "test"
        ret = azureblob._get_containers_cache_filename(container)
        self.assertEqual(
            ret, "/var/cache/salt/master/pillar_azureblob/test-files.cache"
        )

    def test__refresh_containers_cache_file(self):
        """
        Tests the result of _refresh_containers_cache_file to ensure that it successfully copies blob data into a
            cache file.
        """
        blob_client = MockBlobServiceClient()
        container = "test"
        cache_file = tempfile.NamedTemporaryFile()
        with patch.object(
            BlobServiceClient,
            "from_connection_string",
            MagicMock(return_value=blob_client),
        ):
            ret = azureblob._refresh_containers_cache_file(
                "", container, cache_file.name
            )
        cache_file.close()
        self.assertEqual(
            ret,
            {
                "base": {
                    "test": [
                        {
                            "container": None,
                            "name": "test.sls",
                            "prefix": None,
                            "delimiter": "/",
                            "results_per_page": None,
                            "location_mode": None,
                        }
                    ]
                }
            },
        )

    def test__read_containers_cache_file(self):
        """
        Tests the result of _read_containers_cache_file to make sure that it successfully loads in pickled metadata.
        """
        metadata = {
            "base": {
                "test": [
                    {"name": "base/secret.sls", "relevant": "include.sls"},
                    {"name": "blobtest.sls", "irrelevant": "ignore.sls"},
                ]
            }
        }
        cache_file = tempfile.NamedTemporaryFile()
        # Pickles the metadata and stores it in cache_file
        with salt.utils.files.fopen(str(cache_file), "wb") as fp_:
            pickle.dump(metadata, fp_)
        # Checks to see if _read_containers_cache_file can successfully read the pickled metadata from the cache file
        ret = azureblob._read_containers_cache_file(str(cache_file))
        fp_.close()
        os.remove(str(fp_.name))
        cache_file.close()
        self.assertEqual(ret, metadata)

    def test__find_files(self):
        """
        Tests the result of _find_files. Ensures it only finds files and not directories. Ensures it also ignore
            irrelevant files.
        """
        metadata = {
            "test": [
                {"name": "base/secret.sls"},
                {"name": "blobtest.sls", "irrelevant": "ignore.sls"},
                {"name": "base/"},
            ]
        }
        ret = azureblob._find_files(metadata)
        self.assertEqual(ret, {"test": ["base/secret.sls", "blobtest.sls"]})

    def test__find_file_meta1(self):
        """
        Tests the result of _find_file_meta when the metadata contains a blob with the specified path and a blob
            without the specified path.
        """
        metadata = {
            "base": {
                "test": [
                    {"name": "base/secret.sls", "relevant": "include.sls"},
                    {"name": "blobtest.sls", "irrelevant": "ignore.sls"},
                ]
            }
        }
        container = "test"
        saltenv = "base"
        path = "base/secret.sls"
        ret = azureblob._find_file_meta(metadata, container, saltenv, path)
        self.assertEqual(ret, {"name": "base/secret.sls", "relevant": "include.sls"})

    def test__find_file_meta2(self):
        """
        Tests the result of _find_file_meta when the saltenv in metadata does not match the specified saltenv.
        """
        metadata = {"wrong": {"test": [{"name": "base/secret.sls"}]}}
        container = "test"
        saltenv = "base"
        path = "base/secret.sls"
        ret = azureblob._find_file_meta(metadata, container, saltenv, path)
        self.assertEqual(ret, None)

    def test__find_file_meta3(self):
        """
        Tests the result of _find_file_meta when the container in metadata does not match the specified metadata.
        """
        metadata = {"base": {"wrong": [{"name": "base/secret.sls"}]}}
        container = "test"
        saltenv = "base"
        path = "base/secret.sls"
        ret = azureblob._find_file_meta(metadata, container, saltenv, path)
        self.assertEqual(ret, None)



"""
Functions to view the minion's public key information
"""

import os

import salt.utils.crypt


def finger(hash_type=None):
    """
    Return the minion's public key fingerprint

    hash_type
        The hash algorithm used to calculate the fingerprint

    CLI Example:

    .. code-block:: bash

        salt '*' key.finger
    """
    if hash_type is None:
        hash_type = __opts__["hash_type"]

    return salt.utils.crypt.pem_finger(
        os.path.join(__opts__["pki_dir"], "minion.pub"), sum_type=hash_type
    )


def finger_master(hash_type=None):
    """
    Return the fingerprint of the master's public key on the minion.

    hash_type
        The hash algorithm used to calculate the fingerprint

    CLI Example:

    .. code-block:: bash

        salt '*' key.finger_master
    """
    if hash_type is None:
        hash_type = __opts__["hash_type"]

    return salt.utils.crypt.pem_finger(
        os.path.join(__opts__["pki_dir"], "minion_master.pub"), sum_type=hash_type
    )



import salt.modules.win_license as win_license
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class LicenseTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {win_license: {}}

    def test_installed(self):
        """
        Test to see if the given license key is installed
        """
        mock = MagicMock(return_value="Partial Product Key: ABCDE")
        with patch.dict(win_license.__salt__, {"cmd.run": mock}):
            out = win_license.installed("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
            mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /dli")
            self.assertTrue(out)

    def test_installed_diff(self):
        """
        Test to see if the given license key is installed when the key is different
        """
        mock = MagicMock(return_value="Partial Product Key: 12345")
        with patch.dict(win_license.__salt__, {"cmd.run": mock}):
            out = win_license.installed("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
            mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /dli")
            self.assertFalse(out)

    def test_install(self):
        """
        Test installing the given product key
        """
        mock = MagicMock()
        with patch.dict(win_license.__salt__, {"cmd.run": mock}):
            win_license.install("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
            mock.assert_called_once_with(
                r"cscript C:\Windows\System32\slmgr.vbs /ipk "
                "AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE"
            )

    def test_uninstall(self):
        """
        Test uninstalling the given product key
        """
        mock = MagicMock()
        with patch.dict(win_license.__salt__, {"cmd.run": mock}):
            win_license.uninstall()
            mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /upk")

    def test_activate(self):
        """
        Test activating the current product key
        """
        mock = MagicMock()
        with patch.dict(win_license.__salt__, {"cmd.run": mock}):
            win_license.activate()
            mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /ato")

    def test_licensed(self):
        """
        Test checking if the minion is licensed
        """
        mock = MagicMock(return_value="License Status: Licensed")
        with patch.dict(win_license.__salt__, {"cmd.run": mock}):
            win_license.licensed()
            mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /dli")

    def test_info(self):
        """
        Test getting the info about the current license key
        """
        expected = {
            "description": "Prof",
            "licensed": True,
            "name": "Win7",
            "partial_key": "12345",
        }

        mock = MagicMock(
            return_value=(
                "Name: Win7\r\nDescription: Prof\r\nPartial Product Key: 12345\r\n"
                "License Status: Licensed"
            )
        )
        with patch.dict(win_license.__salt__, {"cmd.run": mock}):
            out = win_license.info()
            mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /dli")
            self.assertEqual(out, expected)



"""
tests.integration.master.test_event_return
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This test module is meant to cover the issue being fixed by:

        https://github.com/saltstack/salt/pull/54731
"""

import logging
import os
import shutil
import subprocess
import sys
import time

from salt.utils.nb_popen import NonBlockingPopen
from saltfactories.utils.ports import get_unused_localhost_port
from saltfactories.utils.processes import terminate_process
from tests.support.cli_scripts import ScriptPathMixin
from tests.support.mixins import AdaptedConfigurationTestCaseMixin
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import SkipTest, TestCase

log = logging.getLogger(__name__)


class TestEventReturn(AdaptedConfigurationTestCaseMixin, ScriptPathMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        overrides = {
            "publish_port": get_unused_localhost_port(),
            "ret_port": get_unused_localhost_port(),
            "tcp_master_pub_port": get_unused_localhost_port(),
            "tcp_master_pull_port": get_unused_localhost_port(),
            "tcp_master_publish_pull": get_unused_localhost_port(),
            "tcp_master_workers": get_unused_localhost_port(),
            "runtests_conn_check_port": get_unused_localhost_port(),
            "runtests_log_port": get_unused_localhost_port(),
        }
        overrides["pytest_engine_port"] = overrides["runtests_conn_check_port"]
        temp_config = AdaptedConfigurationTestCaseMixin.get_temp_config(
            "master", **overrides
        )
        cls.root_dir = temp_config["root_dir"]
        cls.config_dir = os.path.dirname(temp_config["conf_file"])
        if temp_config["transport"] == "tcp":
            raise SkipTest("Test only applicable to the ZMQ transport")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.root_dir)
        cls.root_dir = cls.config_dir = None

    def test_master_startup(self):
        proc = NonBlockingPopen(
            [
                sys.executable,
                self.get_script_path("master"),
                "-c",
                RUNTIME_VARS.TMP_CONF_DIR,
                "-l",
                "info",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        out = b""
        err = b""

        # Testing this should never be longer than 1 minute
        max_time = time.time() + 60
        try:
            while True:
                if time.time() > max_time:
                    assert False, "Max timeout ocurred"
                time.sleep(0.5)
                _out = proc.recv()
                _err = proc.recv_err()
                if _out:
                    out += _out
                if _err:
                    err += _err

                if b"DeprecationWarning: object() takes no parameters" in out:
                    self.fail(
                        "'DeprecationWarning: object() takes no parameters' was seen in"
                        " output"
                    )

                if b"TypeError: object() takes no parameters" in out:
                    self.fail(
                        "'TypeError: object() takes no parameters' was seen in output"
                    )

                if b"Setting up the master communication server" in out:
                    # We got past the place we need, stop the process
                    break

                if out is None and err is None:
                    break

                if proc.poll() is not None:
                    break
        finally:
            terminate_process(proc.pid, kill_children=True)



"""
Manage groups on Solaris

.. important::
    If you feel that Salt should be using this module to manage groups on a
    minion, and it is using a different module (or gives an error similar to
    *'group.info' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import logging

import salt.utils.data

log = logging.getLogger(__name__)


try:
    import grp
except ImportError:
    pass

# Define the module's virtual name
__virtualname__ = "group"


def __virtual__():
    """
    Set the group module if the kernel is SunOS
    """
    if __grains__["kernel"] == "SunOS":
        return __virtualname__
    return (
        False,
        "The solaris_group execution module failed to load: "
        "only available on Solaris systems.",
    )


def add(name, gid=None, **kwargs):
    """
    Add the specified group

    CLI Example:

    .. code-block:: bash

        salt '*' group.add foo 3456
    """
    if salt.utils.data.is_true(kwargs.pop("system", False)):
        log.warning("solaris_group module does not support the 'system' argument")
    if kwargs:
        log.warning("Invalid kwargs passed to group.add")

    cmd = "groupadd "
    if gid:
        cmd += "-g {} ".format(gid)
    cmd += name

    ret = __salt__["cmd.run_all"](cmd, python_shell=False)

    return not ret["retcode"]


def delete(name):
    """
    Remove the named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.delete foo
    """
    ret = __salt__["cmd.run_all"]("groupdel {}".format(name), python_shell=False)

    return not ret["retcode"]


def info(name):
    """
    Return information about a group

    CLI Example:

    .. code-block:: bash

        salt '*' group.info foo
    """
    try:
        grinfo = grp.getgrnam(name)
    except KeyError:
        return {}
    else:
        return {
            "name": grinfo.gr_name,
            "passwd": grinfo.gr_passwd,
            "gid": grinfo.gr_gid,
            "members": grinfo.gr_mem,
        }


def getent(refresh=False):
    """
    Return info on all groups

    CLI Example:

    .. code-block:: bash

        salt '*' group.getent
    """
    if "group.getent" in __context__ and not refresh:
        return __context__["group.getent"]

    ret = []
    for grinfo in grp.getgrall():
        ret.append(info(grinfo.gr_name))

    __context__["group.getent"] = ret
    return ret


def chgid(name, gid):
    """
    Change the gid for a named group

    CLI Example:

    .. code-block:: bash

        salt '*' group.chgid foo 4376
    """
    pre_gid = __salt__["file.group_to_gid"](name)
    if gid == pre_gid:
        return True
    cmd = "groupmod -g {} {}".format(gid, name)
    __salt__["cmd.run"](cmd, python_shell=False)
    post_gid = __salt__["file.group_to_gid"](name)
    if post_gid != pre_gid:
        return post_gid == gid
    return False



import copy

from salt.cloud.clouds import qingcloud
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class QingCloudTestCase(TestCase, LoaderModuleMockMixin):
    """
    Unit TestCase for salt.cloud.clouds.qingcloud module.
    """

    def setUp(self):
        self.provider = {
            "providers": {
                "qingcloud": {
                    "qingcloud": {
                        "access_key_id": "key_1234",
                        "secret_access_key": "1234",
                        "zone": "test_zone",
                        "key_filename": "/testfilename",
                        "driver": "qingcloud",
                    }
                }
            }
        }

    def setup_loader_modules(self):
        return {
            qingcloud: {
                "__opts__": {
                    "providers": {"qingcloud": {}},
                    "profiles": {"qingcloud": {}},
                },
                "__active_provider_name__": "qingcloud:qingcloud",
            },
        }

    def test_qingcloud_verify_ssl(self):
        """
        test qinglcoud when using verify_ssl
        """
        patch_sig = patch("salt.cloud.clouds.qingcloud._compute_signature", MagicMock())

        for verify in [True, False, None]:
            mock_requests = MagicMock()
            mock_requests.return_value.status_code = 200
            mock_requests.return_value.text = '{"ret_code": 0}'
            patch_requests = patch("requests.get", mock_requests)
            opts = copy.deepcopy(self.provider)
            opts["providers"]["qingcloud"]["qingcloud"]["verify_ssl"] = verify
            patch_opts = patch.dict(qingcloud.__opts__, opts)
            with patch_sig, patch_requests, patch_opts:
                ret = qingcloud.query()
                self.assertEqual(ret["ret_code"], 0)
                self.assertEqual(
                    mock_requests.call_args_list[0].kwargs["verify"], verify
                )



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_onlyif_req_retcode(state):
    ret = state.single(
        fun="test.succeed_with_changes",
        name="onlyif test",
        onlyif=[{"fun": "test.retcode"}],
    )
    assert ret.result is True
    assert not ret.changes
    assert ret.comment == "onlyif condition is false"

    ret = state.single(
        fun="test.succeed_with_changes",
        name="onlyif test",
        onlyif=[{"fun": "test.retcode", "code": 0}],
    )
    assert ret.result is True
    assert ret.changes
    assert ret.comment == "Success!"



"""
tests.unit.returners.postgres_local_cache_test
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the postgres_local_cache.
"""

import json
import logging
import os
import shutil

import salt.returners.postgres_local_cache as postgres_local_cache
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class PostgresLocalCacheTestCase(TestCase, LoaderModuleMockMixin):
    """
    Tests for the postgre_local_cache returner.
    """

    @classmethod
    def setUpClass(cls):
        cls.TMP_CACHE_DIR = os.path.join(RUNTIME_VARS.TMP, "salt_test_job_cache")
        cls.TMP_JID_DIR = os.path.join(cls.TMP_CACHE_DIR, "jobs")

    def setup_loader_modules(self):
        return {
            postgres_local_cache: {
                "__opts__": {"cachedir": self.TMP_CACHE_DIR, "keep_jobs": 1}
            }
        }

    def tearDown(self):
        """
        Clean up after tests.

        Note that a setUp function is not used in this TestCase because the
        _make_tmp_jid_dirs replaces it.
        """
        if os.path.exists(self.TMP_CACHE_DIR):
            shutil.rmtree(self.TMP_CACHE_DIR)

    def test_returner(self):
        """
        Tests that the returner function
        """
        load = {
            "tgt_type": "glob",
            "fun_args": [],
            "jid": "20200108221839189167",
            "return": True,
            "retcode": 0,
            "success": True,
            "tgt": "minion",
            "cmd": "_return",
            "_stamp": "2020-01-08T22:18:39.197774",
            "arg": [],
            "fun": "test.ping",
            "id": "minion",
        }

        expected = {"return": "True", "retcode": 0, "success": True}

        connect_mock = MagicMock()
        with patch.object(postgres_local_cache, "_get_conn", connect_mock):
            postgres_local_cache.returner(load)

            return_val = None
            for call in connect_mock.mock_calls:
                for arg in call.args:
                    if isinstance(arg, tuple):
                        for val in arg:
                            if isinstance(val, str) and "return" in val:
                                return_val = json.loads(val)

            self.assertIsNotNone(return_val, None)
            self.assertDictEqual(return_val, expected)

    def test_returner_unicode_exception(self):
        """
        Tests that the returner function
        """
        return_val = "Trüe"

        load = {
            "tgt_type": "glob",
            "fun_args": [],
            "jid": "20200108221839189167",
            "return": return_val,
            "retcode": 0,
            "success": True,
            "tgt": "minion",
            "cmd": "_return",
            "_stamp": "2020-01-08T22:18:39.197774",
            "arg": [],
            "fun": "test.ping",
            "id": "minion",
        }

        expected = {"return": "Trüe", "retcode": 0, "success": True}

        connect_mock = MagicMock()
        with patch.object(postgres_local_cache, "_get_conn", connect_mock):
            postgres_local_cache.returner(load)

            return_val = None
            search_string = "return"
            for call in connect_mock.mock_calls:
                for arg in call.args:
                    if isinstance(arg, tuple):
                        for val in arg:
                            if isinstance(val, str):
                                if search_string in val:
                                    return_val = json.loads(val)

            self.assertIsNotNone(return_val, None)
            self.assertDictEqual(return_val, expected)
